# coding: utf8
from typing import Optional

from .. import swap
from ._bermudan_swaption_definition import BermudanSwaptionDefinition
from .._instrument_definition import InstrumentDefinition
from ..swap._swap_definition import SwapInstrumentDefinition
from ._enums import BuySell, CallPut, ExerciseStyle, SwaptionSettlementType


class SwaptionInstrumentDefinition(InstrumentDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported.
    end_date : str, optional
        Expiry date of the option.
    tenor : str, optional
        Tenor of the option.
    notional_amount : float, optional
        The notional amount of the underlying swap.
        By default 1,000,000 is used.
    bermudan_swaption_definition : BermudanSwaptionDefinition, optional

    buy_sell : BuySell
        The direction of the trade in terms of the deal currency : 'Buy' or 'Sell'.
    call_put : CallPut
        Tells if the option is a call or a put.
    exercise_style : ExerciseStyle, optional
        EURO or BERM or AMER.
    settlement_type : SwaptionSettlementType, optional
        The settlement type of the option if the option is exercised.
    underlying_definition : swap.Definition, optional
        The details of the underlying swap.
    spread_vs_atm_in_bp : float, optional
        Spread between strike and atm strike, expressed in basis points (bp).
    strike_percent : float, optional
        strike_percent of the option expressed in % format.
    """

    def __init__(
        self,
        *,
        instrument_tag: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        notional_amount: Optional[float] = None,
        bermudan_swaption_definition: Optional[BermudanSwaptionDefinition] = None,
        buy_sell: Optional[BuySell] = None,
        call_put: Optional[CallPut] = None,
        exercise_style: Optional[ExerciseStyle] = None,
        settlement_type: Optional[SwaptionSettlementType] = None,
        underlying_definition: Optional[swap.Definition] = None,
        spread_vs_atm_in_bp: Optional[float] = None,
        strike_percent: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.instrument_tag = instrument_tag
        self.end_date = end_date
        self.tenor = tenor
        self.notional_amount = notional_amount
        self.bermudan_swaption_definition = bermudan_swaption_definition
        self.buy_sell = buy_sell
        self.call_put = call_put
        self.exercise_style = exercise_style
        self.settlement_type = settlement_type
        self.underlying_definition = underlying_definition
        self.spread_vs_atm_in_bp = spread_vs_atm_in_bp
        self.strike_percent = strike_percent

    @classmethod
    def get_instrument_type(cls):
        return "Swaption"

    @property
    def bermudan_swaption_definition(self):
        """
        :return: object BermudanSwaptionDefinition
        """
        return self._get_object_parameter(
            BermudanSwaptionDefinition, "bermudanSwaptionDefinition"
        )

    @bermudan_swaption_definition.setter
    def bermudan_swaption_definition(self, value):
        self._set_object_parameter(
            BermudanSwaptionDefinition, "bermudanSwaptionDefinition", value
        )

    @property
    def buy_sell(self):
        """
        The side of the deal.
        :return: enum BuySell
        """
        return self._get_enum_parameter(BuySell, "buySell")

    @buy_sell.setter
    def buy_sell(self, value):
        self._set_enum_parameter(BuySell, "buySell", value)

    @property
    def call_put(self):
        """
        Tells if the option is a call or a put.
        :return: enum CallPut
        """
        return self._get_enum_parameter(CallPut, "callPut")

    @call_put.setter
    def call_put(self, value):
        self._set_enum_parameter(CallPut, "callPut", value)

    @property
    def exercise_style(self):
        """
        :return: enum ExerciseStyle
        """
        return self._get_enum_parameter(ExerciseStyle, "exerciseStyle")

    @exercise_style.setter
    def exercise_style(self, value):
        self._set_enum_parameter(ExerciseStyle, "exerciseStyle", value)

    @property
    def settlement_type(self):
        """
        The settlement type of the option if the option is exercised.
        :return: enum SwaptionSettlementType
        """
        return self._get_enum_parameter(SwaptionSettlementType, "settlementType")

    @settlement_type.setter
    def settlement_type(self, value):
        self._set_enum_parameter(SwaptionSettlementType, "settlementType", value)

    @property
    def underlying_definition(self):
        """
        The details of the underlying swap.
        :return: object SwapDefinition
        """
        return self._get_object_parameter(
            SwapInstrumentDefinition, "underlyingDefinition"
        )

    @underlying_definition.setter
    def underlying_definition(self, value):
        self._set_object_parameter(
            SwapInstrumentDefinition, "underlyingDefinition", value
        )

    @property
    def end_date(self):
        """
        Expiry date of the option.
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def instrument_tag(self):
        """
        User defined string to identify the instrument.It can be used to link output results to the instrument definition.
        Only alphabetic, numeric and '- _.#=@' characters are supported.
        :return: str
        """
        return self._get_parameter("instrumentTag")

    @instrument_tag.setter
    def instrument_tag(self, value):
        self._set_parameter("instrumentTag", value)

    @property
    def strike_percent(self):
        """
        StrikePercent of the option expressed in % format.
        :return: float
        """
        return self._get_parameter("strikePercent")

    @strike_percent.setter
    def strike_percent(self, value):
        self._set_parameter("strikePercent", value)

    @property
    def tenor(self):
        """
        Tenor of the option.
        :return: str
        """
        return self._get_parameter("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_parameter("tenor", value)

    @property
    def notional_amount(self):
        """
        The notional amount of the underlying swap.
        By default 1,000,000 is used.
        :return: float
        """
        return self._get_parameter("notionalAmount")

    @notional_amount.setter
    def notional_amount(self, value):
        self._set_parameter("notionalAmount", value)

    @property
    def spread_vs_atm_in_bp(self):
        """
        Spread between strike and atm strike, expressed in basis points (bp).
        :return: float
        """
        return self._get_parameter("spreadVsAtmInBp")

    @spread_vs_atm_in_bp.setter
    def spread_vs_atm_in_bp(self, value):
        self._set_parameter("spreadVsAtmInBp", value)



# coding: utf8

import numpy
from pandas import DataFrame
from pandas import to_numeric

from ..._tools._common import universe_arg_parser, fields_arg_parser
from ...content._content_provider import ErrorParser
from ...delivery._data._data_provider import (
    DataProvider,
    RequestFactory,
    ResponseFactory,
    ContentValidator,
    ValidatorContainer,
)
from ...delivery._stream.stream_cache import StreamCache


# ---------------------------------------------------------------------------
#   Response factory
# ---------------------------------------------------------------------------


class PriceCache:
    def __init__(self, cache: dict):
        self._cache = cache

    def keys(self):
        return self._cache.keys()

    def values(self):
        return self._cache.values()

    def items(self):
        return self._cache.items()

    def __iter__(self):
        return PricingCacheIterator(self)

    def __getitem__(self, name):
        if name in self.keys():
            return self._cache[name]
        raise KeyError(f"{name} not in PriceCache")

    def __len__(self):
        return len(self.keys())

    def __str__(self):
        return str(self._cache)


class PricingCacheIterator:
    def __init__(self, price_cache: PriceCache):
        self._price_cache = price_cache
        self._universe = list(price_cache.keys())
        self._index = 0

    def __next__(self):
        if self._index < len(self._universe):
            name = self._universe[self._index]
            result = self._price_cache[name]
            self._index += 1
            return result
        raise StopIteration()


def create_price_cache(data: dict, fields) -> PriceCache:
    cache = {}
    for item in data:
        key = item.get("Key")
        if key:
            name = key.get("Name")
            service = key.get("Service")
            status = item.get("State")
            cache[name] = StreamCache(
                name=name,
                fields=fields,
                service=service,
                status=status,
                record=item,
            )
    return PriceCache(cache)


def _get_field_value(item_fields, field):
    if item_fields.get(field) is None:
        return numpy.nan
    else:
        return item_fields.get(field)


def _get_field_values(fields, item):
    if item["Type"] == "Status":
        status = numpy.nan
        _code = item["State"]["Code"]
        if _code == "NotEntitled":
            status = "#N/P"
        elif _code == "NotFound":
            status = "#N/F"
        return numpy.full(len(fields), status)
    else:
        return numpy.array([_get_field_value(item["Fields"], f) for f in fields])


def _convert_pricing_json_to_pandas(data, universe, fields):
    data = numpy.array([_get_field_values(fields, item) for item in data])
    df = DataFrame(data, columns=fields, index=universe)
    df = df.apply(to_numeric, errors="ignore")
    if not df.empty:
        df = df.convert_dtypes()
    df = df.reset_index()
    df = df.rename(columns={"index": "Instruments"})
    return df


def pricing_build_df(raw, universe, fields, **kwargs):
    if not fields:
        fields = set()
        for item in raw:
            fields |= item.get("Fields", {}).keys()

        fields = tuple(fields)

    df = _convert_pricing_json_to_pandas(raw, universe, fields)
    return df


class PricingResponseFactory(ResponseFactory):
    def create_success(self, data, *args, **kwargs):
        inst = super().create_success(data, *args, **kwargs)
        content_data = data.get("content_data")
        fields = kwargs.get("fields")
        inst.data.prices = create_price_cache(content_data, fields)
        return inst


# ---------------------------------------------------------------------------
#   Request factory
# ---------------------------------------------------------------------------


class PricingRequestFactory(RequestFactory):
    def get_query_parameters(self, *args, **kwargs) -> list:
        query_parameters = []

        #
        # universe
        #
        universe = kwargs.get("universe")
        if universe:
            universe = universe_arg_parser.get_str(universe, delim=",")
            query_parameters.append(("universe", universe))

        #
        # fields
        #
        fields = kwargs.get("fields")
        if fields:
            fields = fields_arg_parser.get_str(fields, delim=",")
            query_parameters.append(("fields", fields))

        return query_parameters


# ---------------------------------------------------------------------------
#   Content data validator
# ---------------------------------------------------------------------------


class PricingContentValidator(ContentValidator):
    def validate(self, data, *args, **kwargs):
        is_valid = True
        status = data.get("status")

        if status == "Error":
            is_valid = False
            data["error_code"] = data.get("code")
            data["error_message"] = data.get("message")

        return is_valid


# ---------------------------------------------------------------------------
#   Data provider
# ---------------------------------------------------------------------------

pricing_data_provider = DataProvider(
    request=PricingRequestFactory(),
    response=PricingResponseFactory(),
    parser=ErrorParser(),
    validator=ValidatorContainer(content_validator=PricingContentValidator()),
)



from datetime import date, datetime, timedelta
from typing import Optional, Any, Union, List

from ._hp_data_provider import (
    get_content_type_by_interval,
    Adjustments,
    MarketSession,
)
from .._content_provider import ContentProviderLayer
from .._intervals import DayIntervalType, Intervals, get_day_interval_type
from ..._tools import hp_universe_parser, validate_types


class Definition(ContentProviderLayer):
    """
    Summary line of this class that defines parameters for requesting summaries from historical pricing

    Parameters
    ----------
    universe : str or list of str
        The entity universe
    interval : str or Intervals, optional
        The consolidation interval in ISO8601
    start : str or date or datetime or timedelta, optional
        The start date and timestamp of the query in ISO8601 with UTC only
    end : str or date or datetime or timedelta, optional
        The end date and timestamp of the query in ISO8601 with UTC only
    adjustments : list of Adjustments or Adjustments or str, optional
        The adjustment list or Adjustments type
    sessions : list of MarketSession or MarketSession or str, optional
        The list of market session classification or str
    count : int, optional
        The maximum number of data returned. Values range: 1 - 10000
    fields : list, optional
        The list of fields that are to be returned in the response
    closure : Any, optional
        Specifies the parameter that will be merged with the request
    extended_params : dict, optional
        If necessary other parameters

    Examples
    --------
    >>> from refinitiv.data.content.historical_pricing import summaries
    >>> definition_summaries = summaries.Definition("EUR")
    >>> response = definition_summaries.get_data()

    """

    def __init__(
        self,
        universe: Union[str, List[str]],
        interval: Union[str, Intervals] = None,
        start: Optional[Union[str, date, datetime, timedelta]] = None,
        end: Optional[Union[str, date, datetime, timedelta]] = None,
        adjustments: Union[List[Adjustments], Adjustments, str] = None,
        sessions: Union[List[MarketSession], MarketSession, str] = None,
        count: Optional[int] = None,
        fields: Optional[list] = None,
        closure: Optional[Any] = None,
        extended_params: Optional[dict] = None,
    ) -> None:
        # By default, if interval is not defined, interday default value is requested
        day_interval_type = get_day_interval_type(interval or DayIntervalType.INTER)
        content_type = get_content_type_by_interval(day_interval_type)
        universe = hp_universe_parser.get_list(universe)
        validate_types(count, [int, type(None)], "count")

        super().__init__(
            content_type=content_type,
            day_interval_type=day_interval_type,
            universe=universe,
            interval=interval,
            start=start,
            end=end,
            adjustments=adjustments,
            sessions=sessions,
            count=count,
            fields=fields,
            closure=closure,
            extended_params=extended_params,
        )



# coding: utf8

import collections
from enum import Enum, unique
from typing import Any

Error = collections.namedtuple("Error", ["code", "message"])


@unique
class RequestMethod(Enum):
    """
    The RESTful Data service can support multiple methods when
    sending requests to a specified endpoint.
       GET : Request data from the specified endpoint.
       POST : Send data to the specified endpoint to create/update a resource.
       DELETE : Request to delete a resource from a specified endpoint.
       PUT : Send data to the specified endpoint to create/update a resource.
    """

    GET = 1
    POST = 2
    DELETE = 3
    PUT = 4


class EndpointData(object):
    def __init__(self, raw: Any, **kwargs):
        if raw is None:
            raw = {}
        self._raw = raw
        self._kwargs = kwargs

    @property
    def raw(self):
        return self._raw



from typing import Iterable, Union, TYPE_CHECKING

from ..._content_provider import ContentProviderLayer
from ..._content_type import ContentType
from ...._tools import validate_types, validate_bool_value
from .._enums import Package

if TYPE_CHECKING:
    from ....content._types import OptBool, ExtendedParams


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieves recommendation summary.

    Parameters
    ----------
    universe: str, list of str
        The Universe parameter allows the user to define the companies for which the content is returned.

    package: str, Package
        Packages of the content that are subsets in terms of breadth (number of fields) and depth (amount of history) of
        the overall content set. Types of packages: basic, standard, professional

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    extended_params: ExtendedParams, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content import estimates
    >>> definition = estimates.view_summary.recommendations.Definition(universe="IBM.N", package=estimates.Package.BASIC)
    >>> response = definition.get_data()
    """

    def __init__(
        self,
        universe: Union[str, Iterable[str]],
        package: Union[str, Package],
        use_field_names_in_headers: "OptBool" = False,
        extended_params: "ExtendedParams" = None,
    ):
        validate_types(package, [str, Package], "package")
        validate_bool_value(use_field_names_in_headers)

        super().__init__(
            content_type=ContentType.ESTIMATES_VIEW_SUMMARY_RECOMMENDATIONS,
            universe=universe,
            package=package,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        )



# coding: utf8

from enum import Enum, unique

from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class MoneynessType(Enum):
    FWD = "Fwd"
    SIGMA = "Sigma"
    SPOT = "Spot"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(MoneynessType, _MONEYNESS_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_MONEYNESS_TYPE_VALUES_IN_LOWER_BY_MONEYNESS_TYPE, some)


_MONEYNESS_TYPE_VALUES = tuple(t.value for t in MoneynessType)
_MONEYNESS_TYPE_VALUES_IN_LOWER_BY_MONEYNESS_TYPE = {
    name.lower(): item for name, item in MoneynessType.__members__.items()
}



# coding: utf8
import itertools
import re
from typing import Tuple, Any

from .._content_type import ContentType
from ..._core.session._default_session_manager import get_valid_session
from ...delivery._stream import (
    StreamStateEvent,
    StreamStateManager,
    OMMStreamListener,
    _OMMStream,
)
from ...delivery._stream._stream_factory import create_omm_stream
from ...delivery._stream.stream_cache import StreamCache
from ..._tools import cached_property, DEBUG

# regular expression pattern for intra-field position sequence
_partial_update_intra_field_positioning_sequence_regular_expression_pattern = (
    r"[\x1b\x5b|\x9b]([0-9]+)\x60([^\x1b^\x5b|^\x9b]+)"
)
_huge = 1e12


def _decode_intra_field_position_sequence(cached_value, new_value):
    # find all partial update in the value
    tokens = re.findall(
        _partial_update_intra_field_positioning_sequence_regular_expression_pattern,
        new_value,
    )

    # check this value contains a partial update or not?
    if len(tokens) == 0:
        # no partial update required, so done
        return new_value

    # do a partial update
    updated_value = cached_value
    for (offset, replace) in tokens:
        # convert offset from str to int
        offset = int(offset)
        assert offset < len(updated_value)

        # replace the value in the string
        updated_value = (
            updated_value[:offset] + replace + updated_value[offset + len(replace) :]
        )

    # done, return
    return updated_value


_id_iterator = itertools.count()


class StreamingPrice(
    StreamCache, StreamStateManager, OMMStreamListener["StreamingPrice"]
):
    def __init__(
        self,
        name,
        session=None,
        fields=None,
        service=None,
        extended_params=None,
        on_refresh=None,
        on_status=None,
        on_update=None,
        on_complete=None,
        on_error=None,
        parent_id=None,
    ):
        if name is None:
            raise AttributeError("Instrument name must be defined.")

        session = get_valid_session(session)

        StreamCache.__init__(self, name=name, fields=fields, service=service)
        StreamStateManager.__init__(self, logger=session.logger())
        OMMStreamListener.__init__(
            self,
            logger=session.logger(),
            on_refresh=on_refresh,
            on_status=on_status,
            on_update=on_update,
            on_complete=on_complete,
            on_error=on_error,
        )

        self._id = next(_id_iterator)
        if parent_id is not None:
            id_ = f"{parent_id}.{self._id}"
        else:
            id_ = f"{self._id}"

        self._classname: str = f"[{self.__class__.__name__}_{id_} name='{name}']"
        self._session = session
        self._extended_params = extended_params
        self._record = {}

    @cached_property
    def _stream(self) -> _OMMStream:
        omm_stream = create_omm_stream(
            ContentType.STREAMING_PRICING,
            session=self._session,
            name=self._name,
            domain="MarketPrice",
            service=self._service,
            fields=self._fields,
            extended_params=self._extended_params,
            on_refresh=self._on_stream_refresh,
            on_status=self._on_stream_status,
            on_update=self._on_stream_update,
            on_complete=self._on_stream_complete,
            on_error=self._on_stream_error,
        )
        omm_stream.on(StreamStateEvent.CLOSED, self.close)
        return omm_stream

    @property
    def id(self) -> int:
        return self._stream.id

    @property
    def code(self):
        return self._stream.code

    @property
    def message(self):
        return self._stream.message

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.close()

    def _do_pause(self):
        self._stream.pause()

    def _do_resume(self):
        self._stream.resume()

    def _do_close(self, *args, **kwargs):
        id_ = self._stream.id
        self._debug(f"{self._classname} Stop subscription {id_} to {self.name}")
        self._stream.close(*args, **kwargs)

    def _do_open(self, *args, with_updates=True):
        id_ = self._stream.id
        self._debug(f"{self._classname} Open async {id_} to {self.name}")
        self._stream.open(*args, with_updates=with_updates)

    def _decode_partial_update_field(self, key, value):
        """
        This legacy is used to process the partial update
        RETURNS the processed partial update data
        """

        fields = self._record.get("Fields")
        if key not in fields:
            fields[key] = value
            self._warning(f"key {key} not in self._record['Fields']")
            return value

        # process infra-field positioning sequence
        cached_value = fields[key]
        updated_value = _decode_intra_field_position_sequence(cached_value, value)

        # done
        return updated_value

    def _write_to_record(self, message: dict):
        for data in message:
            if data == "Fields":
                fields = message[data]
                # fields data
                # loop over all update items
                for key, value in fields.items():
                    # only string value need to check for a partial update
                    if isinstance(value, str):
                        # value is a string, so check for partial update string
                        # process partial update and update the callback
                        # with processed partial update
                        fields[key] = self._decode_partial_update_field(key, value)

                # update the field data
                self._record.setdefault(data, {})
                self._record[data].update(fields)
            else:
                # not a "Fields" data
                self._record[data] = message[data]

    def _do_on_stream_refresh(self, stream: "_OMMStream", *args) -> Tuple:
        message = args[0]
        self._record = message

        if DEBUG:
            fields = self._record.get("Fields", [])
            num_fields = len(fields)
            self._debug(
                f"|>|>|>|>|>|>{self._classname} "
                f"has fields in record {num_fields} after refresh"
            )

        fields = message.get("Fields")
        return fields

    def _do_on_stream_status(self, stream: "_OMMStream", *args) -> Any:
        message = args[0]
        self._status = message
        return message

    def _do_on_stream_update(self, stream: "_OMMStream", *args) -> Any:
        message = args[0]

        if DEBUG:
            fields = self._record.get("Fields", [])
            num_fields = len(fields)
            self._debug(
                f"|>|>|>|>|>|> {self._classname} "
                f"has fields in record {num_fields} after update"
            )

        self._write_to_record(message)
        fields = message.get("Fields")
        return fields



# coding: utf8

__all__ = ["VolatilityType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class VolatilityType(Enum):
    FLAT = "Flat"
    TERM_STRUCTURE = "TermStructure"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(VolatilityType, _VOLATILITY_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_VOLATILITY_TYPE, some)


_VOLATILITY_TYPE_VALUES = tuple(t.value for t in VolatilityType)
_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_VOLATILITY_TYPE = {
    name.lower(): item for name, item in VolatilityType.__members__.items()
}



from .._content_type import ContentType
from ..._core.session import SessionType, get_valid_session
from ...delivery._data._data_provider import DataProviderLayer

NEWS_UNDERLYING_PLATFORM_KEY = "apis.data.news.underlying-platform"

udf_news_content_type_by_rdp = {
    ContentType.NEWS_STORY_RDP: ContentType.NEWS_STORY_UDF,
    ContentType.NEWS_HEADLINES_RDP: ContentType.NEWS_HEADLINES_UDF,
}


class NewsDataProviderLayer(DataProviderLayer):
    def _check_underlying_platform(self, session):
        underlying_platform = session.config.get(NEWS_UNDERLYING_PLATFORM_KEY) or "rdp"

        if underlying_platform not in {"rdp", "udf"}:
            message = (
                f"Not correct value for '{NEWS_UNDERLYING_PLATFORM_KEY}'. "
                "Possible values: 'udf', 'rdp'"
            )
            session.error(message)
            raise ValueError(message)

        if underlying_platform == "udf":
            if session.type == SessionType.PLATFORM:
                session.debug(
                    "UDF News service cannot be used with platform sessions, RDP News will be used instead. "
                    f"The '{NEWS_UNDERLYING_PLATFORM_KEY}' = 'udf' parameter "
                    "will be discarded, meaning that the regular RDP News service "
                    "will be used for News Story and News Headlines data requests."
                )

            else:
                content_type = udf_news_content_type_by_rdp.get(self._data_type)
                self._initialize(content_type, **self._kwargs)

    def get_data(self, session=None, on_response=None, *args, **kwargs):
        session = get_valid_session(session)
        self._check_underlying_platform(session)
        response = super().get_data(session, on_response, **kwargs)
        return response

    async def get_data_async(self, session=None, on_response=None, *args, **kwargs):
        session = get_valid_session(session)
        self._check_underlying_platform(session)
        response = await super().get_data_async(session, on_response, **kwargs)
        return response



import re
from urllib.parse import ParseResult, urlparse, ParseResultBytes

pattern_1 = re.compile("(.)([A-Z][a-z]+)")
pattern_2 = re.compile("([a-z0-9])([A-Z])")


def camel_to_snake(s):
    if not s:
        return s

    if "_" in s:
        words = [camel_to_snake(w) for w in s.split("_")]
        s = "_".join(words)

    else:
        s = pattern_1.sub(r"\1_\2", s)
        s = pattern_2.sub(r"\1_\2", s)

    return s.lower()


def parse_url(url: str) -> ParseResult:
    import sys

    py_ver = sys.version_info
    if py_ver.major == 3 and py_ver.minor < 9:

        result_urlparse = urlparse(url)

        if isinstance(result_urlparse, ParseResultBytes):
            return result_urlparse

        scheme = result_urlparse.scheme
        netloc = result_urlparse.netloc
        path = result_urlparse.path
        query = result_urlparse.query
        fragment = result_urlparse.fragment

        if not scheme and not netloc and path and ":" in path:
            splitted = path.split(":")
            if len(splitted) == 2:
                scheme, path = splitted

        result = ParseResult(
            scheme=scheme,
            netloc=netloc,
            path=path,
            params=result_urlparse.params,
            query=query,
            fragment=fragment,
        )
    else:

        result = urlparse(url)

    return result


def validate_endpoint_request_url_parameters(url, path_parameters):
    if url == "":
        raise ValueError("Requested URL is missing, please provide valid URL")

    if url.endswith("{universe}") and not path_parameters:
        raise ValueError(
            "Path parameter 'universe' is missing, please provide path parameter"
        )



# coding: utf8

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class AssetClass(Enum):
    FUTURES = "Futures"
    SWAP = "Swap"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(AssetClass, _MAIN_CONSTITUENT_ASSET_CLASS_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _MAIN_CONSTITUENT_ASSET_CLASS_VALUES_IN_LOWER_BY_MAIN_CONSTITUENT_ASSET_CLASS,
            some,
        )


_MAIN_CONSTITUENT_ASSET_CLASS_VALUES = tuple(t.value for t in AssetClass)
_MAIN_CONSTITUENT_ASSET_CLASS_VALUES_IN_LOWER_BY_MAIN_CONSTITUENT_ASSET_CLASS = {
    name.lower(): item for name, item in AssetClass.__members__.items()
}



import asyncio
import threading
from typing import List, Optional

import requests

from ..._open_state import OpenState
from ..._tools import DEBUG

codes = requests.codes

UNAUTHORIZED_CODES = {codes.bad, codes.unauthorized, codes.forbidden}


def is_open(session):
    return session.open_state is OpenState.Opened


def is_closed(session):
    return session.open_state is OpenState.Closed


def loop_exception_handler(loop, context):
    if DEBUG:
        raise context["exception"]


def handle_exception(task):
    exception = None

    try:
        exception = task.exception()
    except asyncio.CancelledError:
        pass

    if exception:
        raise exception


class NullResponse:
    text = ""
    status_code = 0

    def json(self):
        return {}


class Delays:
    def __init__(self, delays: List[int]) -> None:
        self._delays = delays
        self._index = 0

    def next(self) -> int:
        if self._index >= len(self._delays):
            self._index = len(self._delays) - 1
        delay = self._delays[self._index]
        self._index += 1
        return delay

    def reset(self):
        self._index = 0


SECONDS_5 = 5
SECONDS_10 = 10
SECONDS_15 = 15
MINUTE_1 = 60
MINUTES_5 = 5 * MINUTE_1
MINUTES_10 = 10 * MINUTE_1
MINUTES_15 = 15 * MINUTE_1
HOUR_1 = 60 * MINUTE_1
HOURS_2 = 2 * HOUR_1


def get_delays() -> Delays:
    delays = Delays(
        [
            SECONDS_5,
            SECONDS_10,
            SECONDS_15,
            MINUTES_5,
            MINUTES_15,
            HOUR_1,
            HOURS_2,
        ]
    )
    return delays


class Daemon(threading.Thread):
    def __init__(self, interval, name: Optional[str] = None) -> None:
        threading.Thread.__init__(self, name, daemon=True)
        self.finished = threading.Event()
        self.interval = interval

    def cancel(self):
        self.finished.set()

    def run(self):
        while not self.finished.is_set():
            self.finished.wait(self.interval)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.241817

__all__ = ["VolType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class VolType(Enum):
    CALL = "Call"
    PUT = "Put"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(VolType, _VOL_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_VOL_TYPE_VALUES_IN_LOWER_BY_VOL_TYPE, some)


_VOL_TYPE_VALUES = tuple(t.value for t in VolType)
_VOL_TYPE_VALUES_IN_LOWER_BY_VOL_TYPE = {
    name.lower(): item for name, item in VolType.__members__.items()
}



from ....._tools import create_repr

from ..._curves._zc_curve_types import OptConstituents, CurveDefinition, CurveParameters
from ...._content_provider import ContentProviderLayer
from ..._curves._zc_curve_request_item import ZcCurveRequestItem
from ...._content_type import ContentType
from ...._types import OptStr, ExtendedParams


class Definition(ContentProviderLayer):
    """
    Parameters
    ----------
    constituents : Constituents, optional

    curve_definition : ZcCurveDefinitions, optional

    curve_parameters : ZcCurveParameters, optional

    curve_tag : str, optional

    extended_params : dict, optional
        If necessary other parameters.

    Methods
    -------
    get_data(session=session, on_response=on_response)
        Returns a response to the data platform
    get_data_async(session=None, on_response=None, async_mode=None)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> import refinitiv.data.content.ipa.curves.zc_curve as zc_curve
     >>> definition = zc_curve.Definition(
     ...     curve_definition=zc_curve.ZcCurveDefinitions(
     ...         currency="CHF",
     ...         name="CHF LIBOR Swap ZC Curve",
     ...         discounting_tenor="OIS",
     ...     ),
     ...     curve_parameters=zc_curve.ZcCurveParameters(
     ...         use_steps=True
     ...     )
     ... )
     >>> response = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        constituents: OptConstituents = None,
        curve_definition: CurveDefinition = None,
        curve_parameters: CurveParameters = None,
        curve_tag: OptStr = None,
        extended_params: ExtendedParams = None,
    ):
        request_item = ZcCurveRequestItem(
            constituents=constituents,
            curve_definition=curve_definition,
            curve_parameters=curve_parameters,
            curve_tag=curve_tag,
        )
        super().__init__(
            content_type=ContentType.ZC_CURVES,
            universe=request_item,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(self)



# coding: utf8

__all__ = ["DividendType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class DividendType(Enum):
    FORECAST_TABLE = "ForecastTable"
    FORECAST_YIELD = "ForecastYield"
    FUTURES = "Futures"
    HISTORICAL_YIELD = "HistoricalYield"
    IMPLIED_TABLE = "ImpliedTable"
    IMPLIED_YIELD = "ImpliedYield"
    NONE = "None"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(DividendType, _DIVIDEND_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_DIVIDEND_TYPE_VALUES_IN_LOWER_BY_DIVIDEND_TYPE, some)


_DIVIDEND_TYPE_VALUES = tuple(t.value for t in DividendType)
_DIVIDEND_TYPE_VALUES_IN_LOWER_BY_DIVIDEND_TYPE = {
    name.lower(): item for name, item in DividendType.__members__.items()
}



from typing import Iterable, Union, TYPE_CHECKING
from typing import Optional

from ..._content_provider import ContentProviderLayer
from ..._content_type import ContentType
from ...._tools import validate_types, validate_bool_value

if TYPE_CHECKING:
    from ....content._types import ExtendedParams


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve the insider shareholders/declarable stakes investment
    information in the requested company at the specified historical period.

    Parameters
    ----------
    universe: str, list of str
        The Universe parameter allows the user to define the companies for which the content is returned.

    limit: int, optional
        The limit parameter is used for paging. It allows users to select the number of records to be returned.

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    extended_params : ExtendedParams, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content import ownership
    >>> definition = ownership.insider.shareholders_report.Definition("TRI.N")
    >>> response = definition.get_data()
    """

    def __init__(
        self,
        universe: Union[str, Iterable[str]],
        limit: Optional[int] = None,
        use_field_names_in_headers: bool = False,
        extended_params: "ExtendedParams" = None,
    ):
        validate_types(limit, [int, type(None)], "limit")
        validate_bool_value(use_field_names_in_headers)

        super().__init__(
            content_type=ContentType.OWNERSHIP_INSIDER_SHAREHOLDERS_REPORT,
            universe=universe,
            limit=limit,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        )



import re
from datetime import timedelta, datetime, date
from typing import Union, Tuple

import dateutil.parser
from dateutil import tz

ownership_expr = re.compile("[-]?[0-9]+[MDQ]{1}[A]?")


def to_datetime(
    date_value: Union[str, timedelta, Tuple[datetime, date]]
) -> Union[tuple, datetime, None]:
    if date_value is None:
        return None

    if isinstance(date_value, timedelta):
        return datetime.now(tz.tzlocal()) + date_value

    if isinstance(date_value, (datetime, date)):
        return date_value

    try:
        return dateutil.parser.parse(date_value)
    except ValueError as e:
        raise e
    except Exception as e:
        raise ValueError(e)


def _to_utc(datetime_value):
    if datetime_value is None:
        return None
    _value = to_datetime(datetime_value)
    UTC = tz.gettz("UTC")
    _value = _value.astimezone(UTC).replace(tzinfo=None)
    return _value


def to_utc_datetime(datetime_value):
    datetime_value = _to_utc(datetime_value)
    if datetime_value is None:
        return None
    return datetime_value  # .strftime("%Y-%m-%d %H:%M:%S")


def to_utc_date(date_value):
    date_value = _to_utc(date_value)
    if date_value is None:
        return None
    return date_value.date()


def to_utc_datetime_isofmt(datetime_value):
    datetime_value = _to_utc(datetime_value)
    if datetime_value is None:
        return None
    datetime_value = datetime_value.isoformat(timespec="microseconds") + "000Z"
    return datetime_value


def get_date_from_today(days_count):
    if type(days_count) != int:
        raise ValueError(
            "The parameter {} should be an integer, found {}".format(
                days_count, type(days_count)
            )
        )
    return datetime.now(tz.tzlocal()) + timedelta(days=-days_count)


def convert_to_datetime(value, add=None):
    utc = tz.gettz("UTC")

    if value is None:
        return None
    elif isinstance(value, datetime):
        # do nothing
        pass
    elif isinstance(value, timedelta):
        value = datetime.now(tz.tzlocal()).astimezone(utc) + value
    elif isinstance(value, date):
        value = datetime.combine(value, datetime.min.time())
    else:
        try:
            value = dateutil.parser.parse(value)
        except ValueError as e:
            raise e
        except Exception as e:
            raise ValueError(e)
    if add:
        value = value + add
    return value


class Converter:
    def convert(self, value):
        return convert_to_datetime(value)


class Formatter:
    def to_str(self, date_time_value):
        return date_time_value.isoformat(timespec="microseconds")


class CFSFormatter:
    def to_str(self, date_time_value):
        return date_time_value.strftime("%Y-%m-%dT%H:%M:%SZ")


class OwnershipFormatter:
    def to_str(self, date_time_value):
        return date_time_value.strftime("%Y%m%d")


class FundamentalAndReferenceFormatter:
    def to_str(self, date_time):
        s = date_time.isoformat()
        return s.split("T")[0]


class NanosecondsFormatter:
    NANOSECOND_LENGTH = 9

    def to_str(self, date_time):
        s = date_time.isoformat(timespec="microseconds")
        # s => '2021-08-03T10:33:35.554103+00:00000Z'
        s, *_ = s.split("+", maxsplit=1)
        # s => '2021-08-03T10:33:35.554103'
        s, microseconds = s.split(".", maxsplit=1)
        # microseconds => '554103'
        diff = self.NANOSECOND_LENGTH - len(microseconds)
        # nanoseconds => '554103000'
        nanoseconds = f"{microseconds}{diff * '0'}"
        # return => '2021-08-03T10:33:35.554103000Z'
        return f"{s}.{nanoseconds}Z"


class DateTimeAdapter:
    def __init__(self, converter_, formatter_):
        self.converter = converter_
        self.formatter = formatter_

    def convert(self, value):
        return self.converter.convert(value)

    def get_str(self, value):
        return self.formatter.to_str(self.convert(value))


class OwnerShipDateTimeAdapter(DateTimeAdapter):
    def get_str(self, value):
        if isinstance(value, str) and ownership_expr.match(value):
            return value
        return self.formatter.to_str(self.convert(value))


_nanoseconds_formatter = NanosecondsFormatter()
_date_formatter = FundamentalAndReferenceFormatter()
_converter = Converter()

_z_ends_date_time_adapter = DateTimeAdapter(_converter, _nanoseconds_formatter)
_t_ends_date_time_adapter = DateTimeAdapter(_converter, _date_formatter)
hp_datetime_adapter = _z_ends_date_time_adapter
fr_datetime_adapter = _t_ends_date_time_adapter
tds_datetime_adapter = _z_ends_date_time_adapter
custom_insts_datetime_adapter = _z_ends_date_time_adapter
cfs_datetime_adapter = DateTimeAdapter(_converter, CFSFormatter())
ownership_datetime_adapter = OwnerShipDateTimeAdapter(_converter, OwnershipFormatter())



# coding: utf8

from typing import Optional

from .._instrument_pricing_parameters import InstrumentPricingParameters


class PricingParameters(InstrumentPricingParameters):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    market_data_date : str, optional
        The market data date for pricing.
        By default, the market_data_date date is the valuation_date or Today.
    market_value_in_deal_ccy : float, optional
        market_value_in_deal_ccy to override and that will be used as pricing analysis
        input to compute VolatilityPercent. No override is applied by default.
        Note that Premium takes priority over Volatility input.
    nb_iterations : int, optional
        Used for Bermudans and HW1F tree.
    valuation_date : str, optional
        The valuation date for pricing. If not set the valuation date is equal to
        market_data_date or Today. For assets that contains a settlementConvention,
        the default valuation date  is equal to the settlementdate of the Asset that is
        usually the TradeDate+SettlementConvention.

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> rdf.swaption.PricingParameters(valuation_date="2020-04-24", nb_iterations=80)
    """

    def __init__(
        self,
        market_data_date: Optional[str] = None,
        market_value_in_deal_ccy: Optional[float] = None,
        nb_iterations: Optional[int] = None,
        valuation_date: Optional[str] = None,
    ):
        super().__init__()
        self.market_data_date = market_data_date
        self.market_value_in_deal_ccy = market_value_in_deal_ccy
        self.nb_iterations = nb_iterations
        self.valuation_date = valuation_date

    @property
    def market_data_date(self):
        """
        The market data date for pricing.
        By default, the marketDataDate date is the ValuationDate or Today.
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def market_value_in_deal_ccy(self):
        """
        MarketValueInDealCcy to override and that will be used as pricing analysis input to compute VolatilityPercent.
        No override is applied by default. Note that Premium takes priority over Volatility input.
        :return: float
        """
        return self._get_parameter("marketValueInDealCcy")

    @market_value_in_deal_ccy.setter
    def market_value_in_deal_ccy(self, value):
        self._set_parameter("marketValueInDealCcy", value)

    @property
    def nb_iterations(self):
        """
        Used for Bermudans and HW1F tree.
        :return: int
        """
        return self._get_parameter("nbIterations")

    @nb_iterations.setter
    def nb_iterations(self, value):
        self._set_parameter("nbIterations", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.
        If not set the valuation date is equal to MarketDataDate or Today.
        For assets that contains a settlementConvention, the default valuation date  is equal to
        the settlementdate of the Asset that is usually the TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



# coding: utf-8

from ._session_type import SessionType
from ... import _configure as configure


def _retrieve_config_and_set_type(session_name, session_type):
    if session_type == SessionType.PLATFORM:
        session_config = configure.get(
            configure.keys.platform_session(session_name), {}
        )
    elif session_type == SessionType.DESKTOP:
        session_config = configure.get(configure.keys.desktop_session(session_name), {})
    else:
        raise TypeError(
            f"Invalid session type: {session_type}, please set 'desktop' or 'platform'."
        )

    if not session_config:
        raise ValueError(
            f"Can't get config by name: {session_name}. Please check config name"
        )

    return session_config


def _get_session_type_and_name(config_path: str):
    from ._session_provider import get_session_type

    try:
        config_path = config_path or configure.get("sessions.default")
        session_type, session_name = config_path.split(".")

    except ValueError:
        raise ValueError(
            "Please check your 'session_name'. It should be in the following way: 'session_type.session_name'"
        )

    except AttributeError:
        raise AttributeError("Invalid type, please provide string")

    return session_name, get_session_type(session_type)


class Definition(object):
    """
    Create definition object for desktop or platform session.
    When user provides "session-name", then it is checked
    in the config file for matches, if match is found,
    then this session will be taken, otherwise, an exception will be raised

    When user doesn't provide a session name,
    then the default session from the config file
    is taken by using "sessions.default" parameter.

    Parameters
    ----------
        name: str
            name of the session in config. For example: 'platform.default'

    Raises
    ----------
    Exception
        1. If user provided invalid session type in session name.
            Type should be 'platform' or 'desktop'
        2. If app-key not found in config and arguments.

    Examples
    --------
    >>> import refinitiv.data as rd
    >>> platform_session = rd.session.Definition(name="platform.custom-session").get_session()
    """

    def __init__(self, name: str = None) -> None:
        from ._session_provider import _make_session_provider_by_arguments

        self._create_session = _make_session_provider_by_arguments(name)

    def get_session(self):
        session = self._create_session()
        return session



# coding: utf-8

from . import _json_requests as json_requests
from ._tools import (
    is_string_type,
    check_for_string_or_list_of_strings,
    check_for_string,
    get_default_session,
)
import pandas as pd

Symbology_UDF_endpoint = "SymbologySearch"
symbol_types = {
    "ric": "RIC",
    "isin": "ISIN",
    "cusip": "CUSIP",
    "sedol": "SEDOL",
    "ticker": "ticker",
    "lipperid": "lipperID",
    "imo": "IMO",
    "oapermid": "OAPermID",
}


def get_symbology(
    symbol,
    from_symbol_type=None,
    to_symbol_type=None,
    raw_output=False,
    debug=False,
    best_match=True,
):
    """
    Returns a list of instrument names converted into another instrument code.
    For example: convert SEDOL instrument names to RIC names

    Parameters
    ----------
    symbol: string or list of strings
        Single instrument or list of instruments to convert.

    from_symbol_type: string, optional
        Instrument code to convert from.
        Possible values: 'CUSIP', 'ISIN', 'SEDOL', 'RIC', 'ticker', 'lipperID', 'IMO'
        Default: 'RIC'

    to_symbol_type: string or list, optional
        Instrument code to convert to.
        Possible values: 'CUSIP', 'ISIN', 'SEDOL', 'RIC', 'ticker', 'lipperID', 'IMO', 'OAPermID'
        Default: None  (means all symbol types are requested)

    raw_output: boolean, optional
        Set this parameter to True to get the data in json format
        if set to False, the legacy will return a data frame
        Default: False

    debug: boolean, optional
        When set to True, the json request and response are printed.
        Default: False

    best_match: boolean, optional
        When set to True, only primary symbol is requested.
        When set to false, all symbols are requested
        Default: True

    Returns
    -------
    If raw_output is set to True, the data will be returned in the json format.
    If raw_output is False (default value) the data will be returned as a pandas.DataFrame
        The pandas.DataFrame columns are:
        - columns             : Symbol types
        - rows                : Symbol requested
        - cells               : the symbols (None if not found)
        - symbol              : The requested symbol

    Raises
    ------
    Exception
        If request fails or if server returns an error
    ValueError
        If a parameter type or value is wrong

    Examples
    --------
    >> import refinitiv.data as ek
    >> ek.set_app_key('set your app key here')
    >> ISIN_codes = ek.get_symbology(["MSFT.O", "GOOG.O", "IBM.N"], from_symbol_type="RIC", to_symbol_type="ISIN")
    >> ISIN_codes
                    ISIN
    MSFT.O  US5949181045
    GOOG.O  US02079K1079
    IBM.N   US4592001014
    """
    logger = get_default_session().logger()

    if from_symbol_type is None:
        from_symbol_type = "RIC"

    # check if symbol type is string or list of strings
    check_for_string_or_list_of_strings(symbol, "symbol")
    if is_string_type(symbol):
        symbol = [symbol]

    # check if from_symbol type is string
    check_for_string(from_symbol_type, "from_symbol_type")
    try:
        from_symbol_type = symbol_types[from_symbol_type.lower()]
    except Exception:
        error_msg = (
            'from_symbol_type "'
            + from_symbol_type
            + '" should be in '
            + [symbol_types[key] for key in symbol_types].__str__()
        )
        logger.error(error_msg)
        raise ValueError(error_msg)

    # if from_symbol_type is RIC, apply rics = [ric.upper() if ric.islower() else ric for ric in rics ] transformation
    if from_symbol_type == "RIC":
        symbol = [ric.upper() if ric.islower() else ric for ric in symbol]

    # to_symbol_type to None means request all symbol types
    if to_symbol_type is not None:
        # otherwise check if to_symbol type is string or list of strings
        check_for_string_or_list_of_strings(to_symbol_type, "to_symbol_type")
        if is_string_type(to_symbol_type):
            to_symbol_type = [to_symbol_type.strip()]
        try:
            to_symbol_type = [symbol_types[_.lower()] for _ in to_symbol_type]
        except Exception:
            error_msg = (
                "All items in the parameter to_symbol should be in "
                + [symbol_types[key] for key in symbol_types].__str__()
            )
            logger.error(error_msg)
            raise ValueError(error_msg)

    payload = {
        "symbols": symbol,
        "from": from_symbol_type,
        "to": to_symbol_type,
        "bestMatchOnly": best_match,
    }
    response = json_requests.send_json_request(
        Symbology_UDF_endpoint, payload, debug=debug
    )
    result = response.json()
    if raw_output:
        return result
    else:
        if best_match:
            results_dict = dict(
                [(_["symbol"], _["bestMatch"]) for _ in result["mappedSymbols"]]
            )
        else:
            results_dict = dict([(_["symbol"], _) for _ in result["mappedSymbols"]])
        if len(results_dict):
            df_result = pd.DataFrame(results_dict).transpose()
            if not df_result.empty:
                df_result = df_result.convert_dtypes()
        else:
            df_result = pd.DataFrame([])

        return df_result



# coding: utf8

from typing import Optional, Union

from .._instrument_pricing_parameters import InstrumentPricingParameters
from ._enums import (
    DividendType,
    ProjectedIndexCalculationMethod,
    CreditSpreadType,
    PriceSide,
    RedemptionDateType,
    VolatilityType,
    VolatilityTermStructureType,
    BenchmarkYieldSelectionMode,
    YieldType,
    QuoteFallbackLogic,
)
from ._models import BondRoundingParameters


class PricingParameters(InstrumentPricingParameters):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    trade_date : str, optional
        Trade date of the bond. The trade_date is used to compute the default
        valuation_date : By default the rule is that valuation_date = trade_date +
        settlement_convention. Optional. By default, it is equal to market_data_date.
    benchmark_yield_selection_mode : BenchmarkYieldSelectionMode, optional
        The benchmark yield.
        Default value is "Interpolate".
    credit_spread_type : CreditSpreadType, optional
        Credit curve spread type to use during pricing. Applicable for Convertible
        Bonds.
    dividend_type : DividendType, optional
        Underlying stock dividend type used during pricing convertible bond. Applicable
        for Convertible Bonds.
    fx_price_side : PriceSide, optional
        FX price side to consider when retrieving FX rates (Mid, Bid, Ask, Last, Close)
    price_side : PriceSide, optional
        Quoted price side of the bond to use for pricing Analysis: Bid(Bid value),
        Ask(Ask value), Mid(Mid value) Optional. By default the "Mid" price of the bond
        is used.
    projected_index_calculation_method : ProjectedIndexCalculationMethod, optional
        Flag used to define how projected index is computed.
        Default value is "ConstantIndex". It is defaulted to "ForwardIndex"
        for Preferreds and Brazilian Debenture bonds.
    quote_fallback_logic : QuoteFallbackLogic, optional
        Enumeration used to define the fallback logic for the quotation of the
        instrument. Available values are:
        - "None": it means that there's no fallback logic. For example, if the user asks
          for a "Ask" price and instrument is only quoted with a "Bid" price, it is an
          error case.
        - "BestField" : it means that there's a fallback logic to use another market
          data field as quoted price. For example, if the user asks for a "Ask" price
          and instrument is only quoted with a "Bid" price, "Bid" price can be used.
    redemption_date_type : RedemptionDateType, optional
        Redemption type of the bond. It is used to compute the default redemption date:
        - RedemptionAtMaturityDate : yield and price are computed at maturity date.
        - RedemptionAtCallDate : yield and price are computed at call date (next call
          date by default).
        - RedemptionAtPutDate : yield and price are computed at put date (next put date
          by default)..
        - RedemptionAtWorstDate : yield and price are computed at the lowest yield date.
        - RedemptionAtSinkDate : yield and price are computed at sinking fund date.
        - RedemptionAtParDate : yield and price are computed at next par.
        - RedemptionAtPremiumDate : yield and price are computed at next premium.
        - RedemptionAtMakeWholeCallDate : yield and price are computed at Make Whole
          Call date.
        - RedemptionAtAverageLife : yield and price are computed at average life (case
          of sinkable bonds)
        - RedemptionAtNextDate : yield and price are computed at next redemption date
          available.

          Default value is "RedemptionAtWorstDate" for callable
          bond, "RedemptionAtBestDate" for puttable bond or "RedemptionAtMaturityDate".
    rounding_parameters : BondRoundingParameters, optional
        Definition of rounding parameters to be applied on accrued, price or yield.
        By default, rounding parameters are the ones defined in the bond structure.
    volatility_term_structure_type : VolatilityTermStructureType, optional
        Stock volatility trem structure type to use during pricing. Applicable for
        Convertible Bonds.
    volatility_type : VolatilityType, optional
        Volatility type to use during pricing. Applicable for Convertible Bonds.
    yield_type : YieldType, optional
        yield_type that specifies the rate structure.
        The default value is Native.
    adjusted_clean_price : float, optional
        Inflation Adjusted Clean price to override and that will be used as pricing
        analysis input. The currency of the clean price is the cash flow currency (that
        can be different to deal currency especially if "ComputeCashFlowWithReportCcy"
        flag has been set to true). No override is applied by default. Note that only
        one pricing analysis input should be defined.
    adjusted_dirty_price : float, optional
        Inflation Adjusted Dirty price to override and that will be used as pricing
        analysis input. The currency of the dirty price is the cash flow currency (that
        can be different to deal currency especially if "ComputeCashFlowWithReportCcy"
        flag has been set to true). No override is applied by default. Note that only
        one pricing analysis input should be defined.
    adjusted_yield_percent : float, optional
        Inflation Adjusted Yield (expressed in percent) to override and that will be
        used as pricing analysis input. No override is applied by default.
        Note that only one pricing analysis input should be defined.
    apply_tax_to_full_pricing : bool, optional
        Tax Parameters Flag to set these tax parameters for all
        pricing/schedule/risk/spread.
        By default Tax Params are applied only to Muni.
    asset_swap_spread_bp : float, optional
        AssetSwapSpread to override and that will be used as pricing analysis input to
        compute the bond price. No override is applied by default. Note that
        only one pricing anlysis input should be defined.
    benchmark_at_issue_price : float, optional
        Price of benchmark at issue to override and that will be used to compute
        benchmark at redemption spread. No override is applied by default and
        price is computed or retrieved from market data.
    benchmark_at_issue_ric : str, optional
        Ric of benchmark at issue to override and that will be used as pricing analysis
        input to compute the bond price. Optional. No override is applied by default.
        Note that only one pricing analysis input should be defined.
    benchmark_at_issue_spread_bp : float, optional
        Spread of benchmark at issue to override and that will be used as pricing
        analysis input to compute the bond price. No override is applied by
        default. Note that only one pricing analysis input should be defined.
    benchmark_at_issue_yield_percent : float, optional
        Yield of benchmark at issue to override and that will be used to compute
        benchmark at redemption spread. No override is applied by default and
        yield is computed or retrieved from market data.
    benchmark_at_redemption_price : float, optional
        Price of benchmark at redemption to override and that will be used to compute
        benchmark at redemption spread. No override is applied by default and
        price is computed or retrieved from market data.
    benchmark_at_redemption_spread_bp : float, optional
        Spread of benchmark at redemption to override and that will be used as pricing
        analysis input to compute the bond price. No override is applied by
        default. Note that only one pricing analysis input should be defined.
    benchmark_at_redemption_yield_percent : float, optional
        Yield of benchmark at redemption to override and that will be used to compute
        benchmark at redemption spread. No override is applied by default and
        yield is computed or retrieved from market data.
    bond_recovery_rate_percent : float, optional
        Bond Recovery Rate Percent set for convertible bond. Applicable for Convertible
        Bonds.
    cash_amount : float, optional
        Cash amount to override and that will be used as pricing analysis input.
        No override is applied by default. Note that only one pricing analysis
        input should be defined.
    cds_recovery_rate_percent : float, optional
        Recovery rate percent used in credit curve related to convertible. Applicable
        for Convertible Bonds.
    clean_price : float, optional
        Clean price to override and that will be used as pricing analysis input. The
        currency of the clean price is the cash flow currency (that can be different to
        deal currency especially if "ComputeCashFlowWithReportCcy" flag has been set to
        true). No override is applied by default. Note that only one pricing analysis
        input should be defined.
    concession_fee : float, optional
        Fee to apply to the bond price; It is expressed in the same unit that the bond
        price (percent or cash).
    current_yield_percent : float, optional
        Current Yield (expressed in percent) to override and that will be used as
        pricing analysis input. No override is applied by default. Note that
        only one pricing anlysis input should be defined.
    dirty_price : float, optional
        Dirty price to override and that will be used as pricing analysis input. The
        currency of the dirty price is the cash flow currency (that can be different to
        deal currency especially if "ComputeCashFlowWithReportCcy" flag has been set to
        true). No override is applied by default. Note that only one pricing analysis
        input should be defined.
    discount_margin_bp : float, optional
        Discount Margin basis points to override and that will be used as pricing
        analysis input. Available only for Floating Rate Notes. No override is
        applied by default. Note that only one pricing anlysis input should be defined.
    discount_percent : float, optional
        Discount (expressed in percent) to override and that will be used as pricing
        analysis input. Should be used only for bond quoted in discount. Optional. No
        override is applied by default. Note that only one pricing anlysis input should
        be defined.
    dividend_yield_percent : float, optional
        Underlying Stock dividend yield percent. Applicable for Convertible Bonds.
    edsf_benchmark_curve_yield_percent : float, optional
        Yield of Euro-Dollar future benchmark curve (Edsf) to override and that will be
        used to compute Euro-Dollar (Edsf) spread. No override is applied by
        default and yield is computed or retrieved from market data.
    edsf_spread_bp : float, optional
        Spread of Euro-Dollar future benchmark curve (Edsf) to override and that will be
        used as pricing analysis input to compute the bond price. This spread is
        computed for USD Bond whose maturity is under 2 Years. No override is
        applied by default. Note that only one pricing anlysis input should be defined.
    efp_benchmark_price : float, optional
        Price of EFP benchmark to override and that will be used to compute benchmark at
        redemption spread in case the bond is an australian FRN. No override
        is applied by default and price is computed or retrieved from market data.
    efp_benchmark_ric : str, optional
        RIC of EFP benchmark to override and that will be used as pricing analysis input
        to compute the bond price in case the bond is an australian FRN. Ric can be
        only "YTTc1" or "YTCc1".
        Default value is "YTTc1".
    efp_benchmark_yield_percent : float, optional
        Yield of EFP benchmark to override and that will be used to compute benchmark at
        redemption spread in case the bond is an australian FRN. No override
        is applied by default and yield is computed or retrieved from market data.
    efp_spread_bp : float, optional
        Spread of EFP benchmark to override and that will be used as pricing analysis
        input to compute the bond price in case the bond is an australian FRN.
        No override is applied by default. Note that only one pricing analysis input
        should be defined.
    flat_credit_spread_bp : float, optional
        Flat credit spread applied during pricing in basis points. Applicable when
        SpreadType = FlatSpread. Applicable for Convertible Bonds.
    flat_credit_spread_tenor : str, optional
        Flat credit spread tenor on credit curve used during pricing to source credit
        spread value. Applicable for Convertible Bonds.
    fx_stock_correlation : float, optional
        Correlation rate between underlying stock price and FX rate. Applicable for
        cross-currency Convertible Bonds.
    fx_volatility_percent : float, optional
        FX volatility rate percent. Applicable for cross-currency Convertible Bonds.
    fx_volatility_tenor : str, optional
        Tenor on FX volatility to source FX volatility Rate Percent. Applicable for
        cross-currency Convertible Bonds.
    gov_country_benchmark_curve_price : float, optional
        Price of government country benchmark to override and that will be used to
        compute user defined spread. No override is applied by default and price is
        computed or retrieved from market data.
    gov_country_benchmark_curve_yield_percent : float, optional
        Yield of government country benchmark to override and that will be used to
        compute government country spread. No override is applied by default
        and yield is computed or retrieved from market data.
    gov_country_spread_bp : float, optional
        Spread of government country benchmark to override and that will be used as
        pricing analysis input to compute the bond price. Optional. No override is
        applied by default. Note that only one pricing analysis input should be defined.
    government_benchmark_curve_price : float, optional
        Price of government benchmark to override and that will be used to compute user
        defined spread. No override is applied by default and price is
        computed or retrieved from market data.
    government_benchmark_curve_yield_percent : float, optional
        Yield of government benchmark to override and that will be used to compute
        government spread. No override is applied by default and yield is
        computed or retrieved from market data.
    government_spread_bp : float, optional
        Spread of government benchmark to override and that will be used as pricing
        analysis input to compute the bond price. No override is applied by
        default. Note that only one pricing analysis input should be defined.
    issuer_benchmark_curve_yield_percent : float, optional
        Yield of issuer benchmark to override and that will be used to compute issuer
        spread. No override is applied by default and yield is computed or retrieved
        from market data.
    issuer_spread_bp : float, optional
        Spread of issuer benchmark to override and that will be used as pricing analysis
        input to compute the bond price. This spread is computed is for coprorate bonds.
        Optional. No override is applied by default. Note that only one pricing anlysis
        input should be defined.
    market_data_date : str, optional
        The market data date for pricing.
        By default, the market_data_date date is the valuation_date or Today
    market_value_in_deal_ccy : float, optional
        Market value in deal currency. This field can be used to compute notionalAmount
        to apply to get this market value. Optional. By default the value is computed
        from notional amount. NotionalAmount field, market_value_in_deal_ccy field and
        market_value_in_report_ccy field cannot be set at defined at the same time.
    market_value_in_report_ccy : float, optional
        Market value in report currency. This field can be used to compute
        notionalAmount to apply to get this market value. By default the value
        is computed from notional amount. NotionalAmount field, market_value_in_deal_ccy
        field and market_value_in_report_ccy field cannot be set at defined at the same
        time.
    net_price : float, optional
        Net price to override and that will be used as pricing analysis input.
        No override is applied by default. Note that only one pricing anlysis input
        should be defined.
    neutral_yield_percent : float, optional
        Neutral Yield (expressed in percent) to override and that will be used as
        pricing analysis input. This is available only for floating rate notes.
        No override is applied by default. Note that only one pricing analysis
        input should be defined.
    ois_zc_benchmark_curve_yield_percent : float, optional
        Yield of OIS benchmark to override and that will be used to compute OIS spread.
        No override is applied by default and yield is computed or retrieved from market
        data.
    ois_zc_spread_bp : float, optional
        Yield of OIS benchmark to override and that will be used as pricing analysis
        input to compute the bond price. No override is applied by default.
        Note that only one pricing analysis input should be defined.
    option_adjusted_spread_bp : float, optional
        Option Adjusted Spread to override and that will be used as pricing analysis
        input to compute the bond price. No override is applied by default.
        Note that only one pricing analysis input should be defined.
    price : float, optional
        Price to override and that will be used as pricing analysis input. This price
        can be the clean price or dirty price depending on price type defined in bond
        structure. The currency of the price is the cash flow currency (that can be
        different to deal currency especially if "ComputeCashFlowWithReportCcy" flag has
        been set to true). Optional. No override is applied by default. Note that only
        one pricing analysis input should be defined.
    quoted_price : float, optional
        Quoted price to override and that will be used as pricing analysis input. Note
        that a quoted price can be a price, a yield, a discount margin, a spread,...
        depending on quotation type. The currency of the quoted price in case the bonnd
        is price-quoted or cash-quoted is the deal currency (that can be different to
        cash flow currency especially if "ComputeCashFlowWithReportCcy" flag has been
        set to true). No override is applied by default. Note that only one pricing
        analysis input should be defined.
    rating_benchmark_curve_yield_percent : float, optional
        Yield of rating benchmark to override and that will be used to compute rating
        spread. No override is applied by default and yield is computed or retrieved
        from market data.
    rating_spread_bp : float, optional
        Spread of rating benchmark to override and that will be used as pricing analysis
        input to compute the bond price. No override is applied by default.
        Note that only one pricing anlysis input should be defined.
    redemption_date : str, optional
        Redemption date that defines the end date for yield and price computation. Used
        only if redemption date type is set to "RedemptionAtCustomDate"
    sector_rating_benchmark_curve_yield_percent : float, optional
        Yield of sector rating benchmark to override and that will be used to compute
        sector rating spread. No override is applied by default and yield is computed
        or retrieved from market data.
    sector_rating_spread_bp : float, optional
        Spread of sector rating benchmark to override and that will be used as pricing
        analysis input to compute the bond price. No override is applied by default.
        Note that only one pricing anlysis input should be defined.
    settlement_convention : str, optional
        Settlement convention for the bond. By default the rule is that valuation_date =
        trade_date + settlement_convention. By default use the settlement tenor defined
        in the bond structure. Only two parameters among "settlement_convention",
        "market_data_date" and "valuation_date" can be overriden at the same time.
    simple_margin_bp : float, optional
        Simple Margin basis points  to override and that will be used as pricing
        analysis input. Available only for Floating Rate Notes. No override is
        applied by default. Note that only one pricing anlysis input should be defined.
    stock_borrow_rate_percent : float, optional
        Underlying stock borrow rate percent. Applicable for Convertible Bonds.
    stock_flat_volatility_percent : float, optional
        Underlying stock volatility percent used for convertible pricing. Applicable
        when volatility_type = Flat Applicable for Convertible Bonds.
    stock_flat_volatility_tenor : str, optional
        Underlying Stock volatility tenor used during pricing to source volatility
        percent value. Applicable when volatility_type = Flat Applicable for Convertible
        Bonds.
    stock_price_on_default : float, optional
        Assumed stock price agreed in event of default. Applicable for Convertible
        Bonds.
    strip_yield_percent : float, optional
        Strip Yield (expressed in percent) to override and that will be used as pricing
        analysis input. No override is applied by default. Note that only one pricing
        anlysis input should be defined.
    swap_benchmark_curve_yield_percent : float, optional
        Yield of swap benchmark to override and that will be used to compute swap
        spread. No override is applied by default and yield is computed or
        retrieved from market data.
    swap_spread_bp : float, optional
        Spread of swap benchmark to override and that will be used as pricing analysis
        input to compute the bond price. No override is applied by default.
        Note that only one pricing analysis input should be defined.
    tax_on_capital_gain_percent : float, optional
        Tax Rate on capital gain expressed in percent.
        By default no tax is applied that means value is equal to 0.
    tax_on_coupon_percent : float, optional
        Tax Rate on Coupon  expressed in percent.
        By default no tax is applied that means value is equal to 0.
    tax_on_price_percent : float, optional
        Tax Rate on price expressed in percent.
        By default no tax is applied that means value is equal to 0.
    tax_on_yield_percent : float, optional
        Tax Rate on Yield expressed in percent. Also named Tax on Yield Optional.
        By default no tax is applied that means value is equal to 0.
    use_settlement_date_from_quote : bool, optional
        Specify whether to use the settlment date of the quote or the one computed from
        the MarketData Date.
    user_defined_benchmark_price : float, optional
        price of user defined instrument to override and that will be used to compute
        user defined spread. No override is applied by default and price is computed
        or retrieved from market data.
    user_defined_benchmark_yield_percent : float, optional
        Yield of user defined instrument to override and that will be used to compute
        user defined spread. No override is applied by default and yield is computed
        or retrieved from market data.
    user_defined_spread_bp : float, optional
        Spread of user defined instrument to override and that will be used as pricing
        analysis input to compute the bond price. No override is applied by default.
        Note that only one pricing analysis input should be defined.
    valuation_date : str, optional
        The valuation date for pricing. If not set the valuation date is equal
        to market_data_date or Today. For assets that contains a settlement_convention,
        the default valuation date is equal to the settlementdate of the Asset that
        is usually the trade_date+settlement_convention.
    yield_percent : float, optional
        Yield (expressed in percent) to override and that will be used as pricing
        analysis input. No override is applied by default. Note that only one pricing
        analysis input should be defined.
    z_spread_bp : float, optional
        ZSpread to override and that will be used as pricing analysis input to compute
        the bond price. No override is applied by default. Note that only one pricing
        analysis input should be defined.

    Examples
    --------
    >>> import refinitiv.data.content.ipa.financial_contracts as rdf
    >>> definition = rdf.bond.Definition(
    ...    instrument_code="US5YT=RR",
    ...    payment_business_day_convention=rdf.bond.BusinessDayConvention.PREVIOUS_BUSINESS_DAY,
    ...    pricing_parameters=rdf.bond.PricingParameters(
    ...        benchmark_yield_selection_mode=rdf.bond.BenchmarkYieldSelectionMode.INTERPOLATE
    ...    ),
    ...    outputs=[rdf.Outputs.DATA, rdf.Outputs.HEADERS],
    ...    fields=["InstrumentDescription", "MarketDataDate", "Price", "YieldPercent", "ZSpreadBp"]
    ...)
    >>> response = definition.get_data()
    """

    def __init__(
        self,
        trade_date: Optional[str] = None,
        benchmark_yield_selection_mode: Optional[BenchmarkYieldSelectionMode] = None,
        credit_spread_type: Optional[CreditSpreadType] = None,
        dividend_type: Optional[DividendType] = None,
        fx_price_side: Optional[PriceSide] = None,
        price_side: Optional[PriceSide] = None,
        projected_index_calculation_method: Optional[
            ProjectedIndexCalculationMethod
        ] = None,
        quote_fallback_logic: Optional[QuoteFallbackLogic] = None,
        redemption_date_type: Optional[RedemptionDateType] = None,
        rounding_parameters: Union[BondRoundingParameters, dict] = None,
        volatility_term_structure_type: Optional[VolatilityTermStructureType] = None,
        volatility_type: Optional[VolatilityType] = None,
        yield_type: Optional[YieldType] = None,
        adjusted_clean_price: Optional[float] = None,
        adjusted_dirty_price: Optional[float] = None,
        adjusted_yield_percent: Optional[float] = None,
        apply_tax_to_full_pricing: Optional[bool] = None,
        asset_swap_spread_bp: Optional[float] = None,
        benchmark_at_issue_price: Optional[float] = None,
        benchmark_at_issue_ric: Optional[str] = None,
        benchmark_at_issue_spread_bp: Optional[float] = None,
        benchmark_at_issue_yield_percent: Optional[float] = None,
        benchmark_at_redemption_price: Optional[float] = None,
        benchmark_at_redemption_spread_bp: Optional[float] = None,
        benchmark_at_redemption_yield_percent: Optional[float] = None,
        bond_recovery_rate_percent: Optional[float] = None,
        cash_amount: Optional[float] = None,
        cds_recovery_rate_percent: Optional[float] = None,
        clean_price: Optional[float] = None,
        concession_fee: Optional[float] = None,
        current_yield_percent: Optional[float] = None,
        dirty_price: Optional[float] = None,
        discount_margin_bp: Optional[float] = None,
        discount_percent: Optional[float] = None,
        dividend_yield_percent: Optional[float] = None,
        edsf_benchmark_curve_yield_percent: Optional[float] = None,
        edsf_spread_bp: Optional[float] = None,
        efp_benchmark_price: Optional[float] = None,
        efp_benchmark_ric: Optional[str] = None,
        efp_benchmark_yield_percent: Optional[float] = None,
        efp_spread_bp: Optional[float] = None,
        flat_credit_spread_bp: Optional[float] = None,
        flat_credit_spread_tenor: Optional[str] = None,
        fx_stock_correlation: Optional[float] = None,
        fx_volatility_percent: Optional[float] = None,
        fx_volatility_tenor: Optional[str] = None,
        gov_country_benchmark_curve_price: Optional[float] = None,
        gov_country_benchmark_curve_yield_percent: Optional[float] = None,
        gov_country_spread_bp: Optional[float] = None,
        government_benchmark_curve_price: Optional[float] = None,
        government_benchmark_curve_yield_percent: Optional[float] = None,
        government_spread_bp: Optional[float] = None,
        issuer_benchmark_curve_yield_percent: Optional[float] = None,
        issuer_spread_bp: Optional[float] = None,
        market_data_date: Optional[str] = None,
        market_value_in_deal_ccy: Optional[float] = None,
        market_value_in_report_ccy: Optional[float] = None,
        net_price: Optional[float] = None,
        neutral_yield_percent: Optional[float] = None,
        ois_zc_benchmark_curve_yield_percent: Optional[float] = None,
        ois_zc_spread_bp: Optional[float] = None,
        option_adjusted_spread_bp: Optional[float] = None,
        price: Optional[float] = None,
        quoted_price: Optional[float] = None,
        rating_benchmark_curve_yield_percent: Optional[float] = None,
        rating_spread_bp: Optional[float] = None,
        redemption_date: Optional[str] = None,
        sector_rating_benchmark_curve_yield_percent: Optional[float] = None,
        sector_rating_spread_bp: Optional[float] = None,
        settlement_convention: Optional[str] = None,
        simple_margin_bp: Optional[float] = None,
        stock_borrow_rate_percent: Optional[float] = None,
        stock_flat_volatility_percent: Optional[float] = None,
        stock_flat_volatility_tenor: Optional[str] = None,
        stock_price_on_default: Optional[float] = None,
        strip_yield_percent: Optional[float] = None,
        swap_benchmark_curve_yield_percent: Optional[float] = None,
        swap_spread_bp: Optional[float] = None,
        tax_on_capital_gain_percent: Optional[float] = None,
        tax_on_coupon_percent: Optional[float] = None,
        tax_on_price_percent: Optional[float] = None,
        tax_on_yield_percent: Optional[float] = None,
        use_settlement_date_from_quote: Optional[bool] = None,
        user_defined_benchmark_price: Optional[float] = None,
        user_defined_benchmark_yield_percent: Optional[float] = None,
        user_defined_spread_bp: Optional[float] = None,
        valuation_date: Optional[str] = None,
        yield_percent: Optional[float] = None,
        z_spread_bp: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.trade_date = trade_date
        self.benchmark_yield_selection_mode = benchmark_yield_selection_mode
        self.credit_spread_type = credit_spread_type
        self.dividend_type = dividend_type
        self.fx_price_side = fx_price_side
        self.price_side = price_side
        self.projected_index_calculation_method = projected_index_calculation_method
        self.quote_fallback_logic = quote_fallback_logic
        self.redemption_date_type = redemption_date_type
        self.rounding_parameters = rounding_parameters
        self.volatility_term_structure_type = volatility_term_structure_type
        self.volatility_type = volatility_type
        self.yield_type = yield_type
        self.adjusted_clean_price = adjusted_clean_price
        self.adjusted_dirty_price = adjusted_dirty_price
        self.adjusted_yield_percent = adjusted_yield_percent
        self.apply_tax_to_full_pricing = apply_tax_to_full_pricing
        self.asset_swap_spread_bp = asset_swap_spread_bp
        self.benchmark_at_issue_price = benchmark_at_issue_price
        self.benchmark_at_issue_ric = benchmark_at_issue_ric
        self.benchmark_at_issue_spread_bp = benchmark_at_issue_spread_bp
        self.benchmark_at_issue_yield_percent = benchmark_at_issue_yield_percent
        self.benchmark_at_redemption_price = benchmark_at_redemption_price
        self.benchmark_at_redemption_spread_bp = benchmark_at_redemption_spread_bp
        self.benchmark_at_redemption_yield_percent = (
            benchmark_at_redemption_yield_percent
        )
        self.bond_recovery_rate_percent = bond_recovery_rate_percent
        self.cash_amount = cash_amount
        self.cds_recovery_rate_percent = cds_recovery_rate_percent
        self.clean_price = clean_price
        self.concession_fee = concession_fee
        self.current_yield_percent = current_yield_percent
        self.dirty_price = dirty_price
        self.discount_margin_bp = discount_margin_bp
        self.discount_percent = discount_percent
        self.dividend_yield_percent = dividend_yield_percent
        self.edsf_benchmark_curve_yield_percent = edsf_benchmark_curve_yield_percent
        self.edsf_spread_bp = edsf_spread_bp
        self.efp_benchmark_price = efp_benchmark_price
        self.efp_benchmark_ric = efp_benchmark_ric
        self.efp_benchmark_yield_percent = efp_benchmark_yield_percent
        self.efp_spread_bp = efp_spread_bp
        self.flat_credit_spread_bp = flat_credit_spread_bp
        self.flat_credit_spread_tenor = flat_credit_spread_tenor
        self.fx_stock_correlation = fx_stock_correlation
        self.fx_volatility_percent = fx_volatility_percent
        self.fx_volatility_tenor = fx_volatility_tenor
        self.gov_country_benchmark_curve_price = gov_country_benchmark_curve_price
        self.gov_country_benchmark_curve_yield_percent = (
            gov_country_benchmark_curve_yield_percent
        )
        self.gov_country_spread_bp = gov_country_spread_bp
        self.government_benchmark_curve_price = government_benchmark_curve_price
        self.government_benchmark_curve_yield_percent = (
            government_benchmark_curve_yield_percent
        )
        self.government_spread_bp = government_spread_bp
        self.issuer_benchmark_curve_yield_percent = issuer_benchmark_curve_yield_percent
        self.issuer_spread_bp = issuer_spread_bp
        self.market_data_date = market_data_date
        self.market_value_in_deal_ccy = market_value_in_deal_ccy
        self.market_value_in_report_ccy = market_value_in_report_ccy
        self.net_price = net_price
        self.neutral_yield_percent = neutral_yield_percent
        self.ois_zc_benchmark_curve_yield_percent = ois_zc_benchmark_curve_yield_percent
        self.ois_zc_spread_bp = ois_zc_spread_bp
        self.option_adjusted_spread_bp = option_adjusted_spread_bp
        self.price = price
        self.quoted_price = quoted_price
        self.rating_benchmark_curve_yield_percent = rating_benchmark_curve_yield_percent
        self.rating_spread_bp = rating_spread_bp
        self.redemption_date = redemption_date
        self.sector_rating_benchmark_curve_yield_percent = (
            sector_rating_benchmark_curve_yield_percent
        )
        self.sector_rating_spread_bp = sector_rating_spread_bp
        self.settlement_convention = settlement_convention
        self.simple_margin_bp = simple_margin_bp
        self.stock_borrow_rate_percent = stock_borrow_rate_percent
        self.stock_flat_volatility_percent = stock_flat_volatility_percent
        self.stock_flat_volatility_tenor = stock_flat_volatility_tenor
        self.stock_price_on_default = stock_price_on_default
        self.strip_yield_percent = strip_yield_percent
        self.swap_benchmark_curve_yield_percent = swap_benchmark_curve_yield_percent
        self.swap_spread_bp = swap_spread_bp
        self.tax_on_capital_gain_percent = tax_on_capital_gain_percent
        self.tax_on_coupon_percent = tax_on_coupon_percent
        self.tax_on_price_percent = tax_on_price_percent
        self.tax_on_yield_percent = tax_on_yield_percent
        self.use_settlement_date_from_quote = use_settlement_date_from_quote
        self.user_defined_benchmark_price = user_defined_benchmark_price
        self.user_defined_benchmark_yield_percent = user_defined_benchmark_yield_percent
        self.user_defined_spread_bp = user_defined_spread_bp
        self.valuation_date = valuation_date
        self.yield_percent = yield_percent
        self.z_spread_bp = z_spread_bp

    @property
    def benchmark_yield_selection_mode(self):
        """
        The benchmark yield selection mode:
        - Interpolate : do an interpolatation on yield curve to compute the reference
          yield.
        - Nearest : use the nearest point to find the reference yield. Optional. Default
          value is "Interpolate".
        :return: enum BenchmarkYieldSelectionMode
        """
        return self._get_enum_parameter(
            BenchmarkYieldSelectionMode, "benchmarkYieldSelectionMode"
        )

    @benchmark_yield_selection_mode.setter
    def benchmark_yield_selection_mode(self, value):
        self._set_enum_parameter(
            BenchmarkYieldSelectionMode, "benchmarkYieldSelectionMode", value
        )

    @property
    def credit_spread_type(self):
        """
        Credit curve spread type to use during pricing. Applicable for Convertible
        Bonds.
        :return: enum CreditSpreadType
        """
        return self._get_enum_parameter(CreditSpreadType, "creditSpreadType")

    @credit_spread_type.setter
    def credit_spread_type(self, value):
        self._set_enum_parameter(CreditSpreadType, "creditSpreadType", value)

    @property
    def dividend_type(self):
        """
        Underlying stock dividend type used during pricing convertible bond. Applicable
        for Convertible Bonds.
        :return: enum DividendType
        """
        return self._get_enum_parameter(DividendType, "dividendType")

    @dividend_type.setter
    def dividend_type(self, value):
        self._set_enum_parameter(DividendType, "dividendType", value)

    @property
    def fx_price_side(self):
        """
        FX price side to consider when retrieving FX rates (Mid, Bid, Ask, Last, Close)
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "fxPriceSide")

    @fx_price_side.setter
    def fx_price_side(self, value):
        self._set_enum_parameter(PriceSide, "fxPriceSide", value)

    @property
    def price_side(self):
        """
        Quoted price side of the bond to use for pricing Analysis: Bid(Bid value),
        Ask(Ask value), Mid(Mid value) Optional. By default the "Mid" price of the bond
        is used.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def projected_index_calculation_method(self):
        """
        Flag used to define how projected index is computed. Avalaible values are:
        - "ConstantIndex" : future index values are considered as constant and equal to
          projected index value.
        - "ForwardIndex" : future index values are computed using a forward curve.
          Optional. Default value is "ConstantIndex". It is defaulted to "ForwardIndex"
          for Preferreds and Brazilian Debenture bonds.
        :return: enum ProjectedIndexCalculationMethod
        """
        return self._get_enum_parameter(
            ProjectedIndexCalculationMethod, "projectedIndexCalculationMethod"
        )

    @projected_index_calculation_method.setter
    def projected_index_calculation_method(self, value):
        self._set_enum_parameter(
            ProjectedIndexCalculationMethod, "projectedIndexCalculationMethod", value
        )

    @property
    def quote_fallback_logic(self):
        """
        Enumeration used to define the fallback logic for the quotation of the
        instrument. Available values are:
        - "None": it means that there's no fallback logic. For example, if the user asks
          for a "Ask" price and instrument is only quoted with a "Bid" price, it is an
          error case.
        - "BestField" : it means that there's a fallback logic to use another market
          data field as quoted price. For example, if the user asks for a "Ask" price
          and instrument is only quoted with a "Bid" price, "Bid" price can be used.
        :return: enum QuoteFallbackLogic
        """
        return self._get_enum_parameter(QuoteFallbackLogic, "quoteFallbackLogic")

    @quote_fallback_logic.setter
    def quote_fallback_logic(self, value):
        self._set_enum_parameter(QuoteFallbackLogic, "quoteFallbackLogic", value)

    @property
    def redemption_date_type(self):
        """
        Redemption type of the bond. It is used to compute the default redemption date:
        - RedemptionAtMaturityDate : yield and price are computed at maturity date.
        - RedemptionAtCallDate : yield and price are computed at call date (next call
          date by default).
        - RedemptionAtPutDate : yield and price are computed at put date (next put date
          by default)..
        - RedemptionAtWorstDate : yield and price are computed at the lowest yield date.
        - RedemptionAtSinkDate : yield and price are computed at sinking fund date.
        - RedemptionAtParDate : yield and price are computed at next par.
        - RedemptionAtPremiumDate : yield and price are computed at next premium.
        - RedemptionAtMakeWholeCallDate : yield and price are computed at Make Whole
          Call date.
        - RedemptionAtAverageLife : yield and price are computed at average life (case
          of sinkable bonds)
        - RedemptionAtNextDate : yield and price are computed at next redemption date
          available. Optional. Default value is "RedemptionAtWorstDate" for callable
          bond, "RedemptionAtBestDate" for puttable bond or "RedemptionAtMaturityDate".
        :return: enum RedemptionDateType
        """
        return self._get_enum_parameter(RedemptionDateType, "redemptionDateType")

    @redemption_date_type.setter
    def redemption_date_type(self, value):
        self._set_enum_parameter(RedemptionDateType, "redemptionDateType", value)

    @property
    def rounding_parameters(self):
        """
        Definition of rounding parameters to be applied on accrued, price or yield.
        Optional. By default, rounding parameters are the ones defined in the bond
        structure.
        :return: object BondRoundingParameters
        """
        return self._get_object_parameter(BondRoundingParameters, "roundingParameters")

    @rounding_parameters.setter
    def rounding_parameters(self, value):
        self._set_object_parameter(BondRoundingParameters, "roundingParameters", value)

    @property
    def volatility_term_structure_type(self):
        """
        Stock volatility trem structure type to use during pricing. Applicable for
        Convertible Bonds.
        :return: enum VolatilityTermStructureType
        """
        return self._get_enum_parameter(
            VolatilityTermStructureType, "volatilityTermStructureType"
        )

    @volatility_term_structure_type.setter
    def volatility_term_structure_type(self, value):
        self._set_enum_parameter(
            VolatilityTermStructureType, "volatilityTermStructureType", value
        )

    @property
    def volatility_type(self):
        """
        Volatility type to use during pricing. Applicable for Convertible Bonds.
        :return: enum VolatilityType
        """
        return self._get_enum_parameter(VolatilityType, "volatilityType")

    @volatility_type.setter
    def volatility_type(self, value):
        self._set_enum_parameter(VolatilityType, "volatilityType", value)

    @property
    def yield_type(self):
        """
        YieldType that specifies the rate structure.
        - Native : no specific yield type is defined.
        - UsGovt_Actual_Actual_6M : US Govt Act/Act 6M YTA.
        - Isma_30_360_6M : ISMA 30/360 6M YTA.
        - Euroland_Actual_Actual_6M : Euroland Equivalent Act/Act 6M YTA.
        - Money_Market_Actual_360_6M : Money Market Act/360 6M YTA.
        - Money_Market_Actual_365_6M : Money Market Act/365 6M YTA.
        - Money_Market_Actual_Actual_6M : Money Market Act/Act 6M YTA.
        - Bond_Actual_364_6M : Bond Market Act/364 6M YTA.
        - Japanese_Simple_JAP_6M : Japanese Simple JAP 6M YTA.
        - Japanese_Compunded_30_360_6M : Japanese Compounded 30/360 6M YTA.
        - Moosmueller_30_360_6M : Moosmueller 30/360 6M YTA.
        - Braess_Frangmeyer_30_360_6M : Braess-Frangmeyer 30/360 6M YTA.
        - Weekend_30_360 : Week End 30/360 6M YTA Optional. The default value is Native.
        :return: enum YieldType
        """
        return self._get_enum_parameter(YieldType, "yieldType")

    @yield_type.setter
    def yield_type(self, value):
        self._set_enum_parameter(YieldType, "yieldType", value)

    @property
    def adjusted_clean_price(self):
        """
        Inflation Adjusted Clean price to override and that will be used as pricing
        analysis input. The currency of the clean price is the cash flow currency (that
        can be different to deal currency especially if "ComputeCashFlowWithReportCcy"
        flag has been set to true). Optional. No override is applied by default. Note
        that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("adjustedCleanPrice")

    @adjusted_clean_price.setter
    def adjusted_clean_price(self, value):
        self._set_parameter("adjustedCleanPrice", value)

    @property
    def adjusted_dirty_price(self):
        """
        Inflation Adjusted Dirty price to override and that will be used as pricing
        analysis input. The currency of the dirty price is the cash flow currency (that
        can be different to deal currency especially if "ComputeCashFlowWithReportCcy"
        flag has been set to true). Optional. No override is applied by default. Note
        that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("adjustedDirtyPrice")

    @adjusted_dirty_price.setter
    def adjusted_dirty_price(self, value):
        self._set_parameter("adjustedDirtyPrice", value)

    @property
    def adjusted_yield_percent(self):
        """
        Inflation Adjusted Yield (expressed in percent) to override and that will be
        used as pricing analysis input. Optional. No override is applied by default.
        Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("adjustedYieldPercent")

    @adjusted_yield_percent.setter
    def adjusted_yield_percent(self, value):
        self._set_parameter("adjustedYieldPercent", value)

    @property
    def apply_tax_to_full_pricing(self):
        """
        Tax Parameters Flag to set these tax parameters for all
        pricing/schedule/risk/spread. Optional. By default Tax Params are applied only
        to Muni.
        :return: bool
        """
        return self._get_parameter("applyTaxToFullPricing")

    @apply_tax_to_full_pricing.setter
    def apply_tax_to_full_pricing(self, value):
        self._set_parameter("applyTaxToFullPricing", value)

    @property
    def asset_swap_spread_bp(self):
        """
        AssetSwapSpread to override and that will be used as pricing analysis input to
        compute the bond price. Optional. No override is applied by default. Note that
        only one pricing anlysis input should be defined.
        :return: float
        """
        return self._get_parameter("assetSwapSpreadBp")

    @asset_swap_spread_bp.setter
    def asset_swap_spread_bp(self, value):
        self._set_parameter("assetSwapSpreadBp", value)

    @property
    def benchmark_at_issue_price(self):
        """
        Price of benchmark at issue to override and that will be used to compute
        benchmark at redemption spread. Optional. No override is applied by default and
        price is computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("benchmarkAtIssuePrice")

    @benchmark_at_issue_price.setter
    def benchmark_at_issue_price(self, value):
        self._set_parameter("benchmarkAtIssuePrice", value)

    @property
    def benchmark_at_issue_ric(self):
        """
        Ric of benchmark at issue to override and that will be used as pricing analysis
        input to compute the bond price. Optional. No override is applied by default.
        Note that only one pricing analysis input should be defined.
        :return: str
        """
        return self._get_parameter("benchmarkAtIssueRic")

    @benchmark_at_issue_ric.setter
    def benchmark_at_issue_ric(self, value):
        self._set_parameter("benchmarkAtIssueRic", value)

    @property
    def benchmark_at_issue_spread_bp(self):
        """
        Spread of benchmark at issue to override and that will be used as pricing
        analysis input to compute the bond price. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("benchmarkAtIssueSpreadBp")

    @benchmark_at_issue_spread_bp.setter
    def benchmark_at_issue_spread_bp(self, value):
        self._set_parameter("benchmarkAtIssueSpreadBp", value)

    @property
    def benchmark_at_issue_yield_percent(self):
        """
        Yield of benchmark at issue to override and that will be used to compute
        benchmark at redemption spread. Optional. No override is applied by default and
        yield is computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("benchmarkAtIssueYieldPercent")

    @benchmark_at_issue_yield_percent.setter
    def benchmark_at_issue_yield_percent(self, value):
        self._set_parameter("benchmarkAtIssueYieldPercent", value)

    @property
    def benchmark_at_redemption_price(self):
        """
        Price of benchmark at redemption to override and that will be used to compute
        benchmark at redemption spread. Optional. No override is applied by default and
        price is computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("benchmarkAtRedemptionPrice")

    @benchmark_at_redemption_price.setter
    def benchmark_at_redemption_price(self, value):
        self._set_parameter("benchmarkAtRedemptionPrice", value)

    @property
    def benchmark_at_redemption_spread_bp(self):
        """
        Spread of benchmark at redemption to override and that will be used as pricing
        analysis input to compute the bond price. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("benchmarkAtRedemptionSpreadBp")

    @benchmark_at_redemption_spread_bp.setter
    def benchmark_at_redemption_spread_bp(self, value):
        self._set_parameter("benchmarkAtRedemptionSpreadBp", value)

    @property
    def benchmark_at_redemption_yield_percent(self):
        """
        Yield of benchmark at redemption to override and that will be used to compute
        benchmark at redemption spread. Optional. No override is applied by default and
        yield is computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("benchmarkAtRedemptionYieldPercent")

    @benchmark_at_redemption_yield_percent.setter
    def benchmark_at_redemption_yield_percent(self, value):
        self._set_parameter("benchmarkAtRedemptionYieldPercent", value)

    @property
    def bond_recovery_rate_percent(self):
        """
        Bond Recovery Rate Percent set for convertible bond. Applicable for Convertible
        Bonds.
        :return: float
        """
        return self._get_parameter("bondRecoveryRatePercent")

    @bond_recovery_rate_percent.setter
    def bond_recovery_rate_percent(self, value):
        self._set_parameter("bondRecoveryRatePercent", value)

    @property
    def cash_amount(self):
        """
        Cash amount to override and that will be used as pricing analysis input.
        Optional. No override is applied by default. Note that only one pricing analysis
        input should be defined.
        :return: float
        """
        return self._get_parameter("cashAmount")

    @cash_amount.setter
    def cash_amount(self, value):
        self._set_parameter("cashAmount", value)

    @property
    def cds_recovery_rate_percent(self):
        """
        Recovery rate percent used in credit curve related to convertible. Applicable
        for Convertible Bonds.
        :return: float
        """
        return self._get_parameter("cdsRecoveryRatePercent")

    @cds_recovery_rate_percent.setter
    def cds_recovery_rate_percent(self, value):
        self._set_parameter("cdsRecoveryRatePercent", value)

    @property
    def clean_price(self):
        """
        Clean price to override and that will be used as pricing analysis input. The
        currency of the clean price is the cash flow currency (that can be different to
        deal currency especially if "ComputeCashFlowWithReportCcy" flag has been set to
        true). Optional. No override is applied by default. Note that only one pricing
        analysis input should be defined.
        :return: float
        """
        return self._get_parameter("cleanPrice")

    @clean_price.setter
    def clean_price(self, value):
        self._set_parameter("cleanPrice", value)

    @property
    def concession_fee(self):
        """
        Fee to apply to the bond price; It is expressed in the same unit that the bond
        price (percent or cash).
        :return: float
        """
        return self._get_parameter("concessionFee")

    @concession_fee.setter
    def concession_fee(self, value):
        self._set_parameter("concessionFee", value)

    @property
    def current_yield_percent(self):
        """
        Current Yield (expressed in percent) to override and that will be used as
        pricing analysis input. Optional. No override is applied by default. Note that
        only one pricing anlysis input should be defined.
        :return: float
        """
        return self._get_parameter("currentYieldPercent")

    @current_yield_percent.setter
    def current_yield_percent(self, value):
        self._set_parameter("currentYieldPercent", value)

    @property
    def dirty_price(self):
        """
        Dirty price to override and that will be used as pricing analysis input. The
        currency of the dirty price is the cash flow currency (that can be different to
        deal currency especially if "ComputeCashFlowWithReportCcy" flag has been set to
        true). Optional. No override is applied by default. Note that only one pricing
        analysis input should be defined.
        :return: float
        """
        return self._get_parameter("dirtyPrice")

    @dirty_price.setter
    def dirty_price(self, value):
        self._set_parameter("dirtyPrice", value)

    @property
    def discount_margin_bp(self):
        """
        Discount Margin basis points to override and that will be used as pricing
        analysis input. Available only for Floating Rate Notes. Optional.No override is
        applied by default. Note that only one pricing anlysis input should be defined.
        :return: float
        """
        return self._get_parameter("discountMarginBp")

    @discount_margin_bp.setter
    def discount_margin_bp(self, value):
        self._set_parameter("discountMarginBp", value)

    @property
    def discount_percent(self):
        """
        Discount (expressed in percent) to override and that will be used as pricing
        analysis input. Should be used only for bond quoted in discount. Optional. No
        override is applied by default. Note that only one pricing anlysis input should
        be defined.
        :return: float
        """
        return self._get_parameter("discountPercent")

    @discount_percent.setter
    def discount_percent(self, value):
        self._set_parameter("discountPercent", value)

    @property
    def dividend_yield_percent(self):
        """
        Underlying Stock dividend yield percent. Applicable for Convertible Bonds.
        :return: float
        """
        return self._get_parameter("dividendYieldPercent")

    @dividend_yield_percent.setter
    def dividend_yield_percent(self, value):
        self._set_parameter("dividendYieldPercent", value)

    @property
    def edsf_benchmark_curve_yield_percent(self):
        """
        Yield of Euro-Dollar future benchmark curve (Edsf) to override and that will be
        used to compute Euro-Dollar (Edsf) spread. Optional. No override is applied by
        default and yield is computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("edsfBenchmarkCurveYieldPercent")

    @edsf_benchmark_curve_yield_percent.setter
    def edsf_benchmark_curve_yield_percent(self, value):
        self._set_parameter("edsfBenchmarkCurveYieldPercent", value)

    @property
    def edsf_spread_bp(self):
        """
        Spread of Euro-Dollar future benchmark curve (Edsf) to override and that will be
        used as pricing analysis input to compute the bond price. This spread is
        computed for USD Bond whose maturity is under 2 Years. Optional. No override is
        applied by default. Note that only one pricing anlysis input should be defined.
        :return: float
        """
        return self._get_parameter("edsfSpreadBp")

    @edsf_spread_bp.setter
    def edsf_spread_bp(self, value):
        self._set_parameter("edsfSpreadBp", value)

    @property
    def efp_benchmark_price(self):
        """
        Price of EFP benchmark to override and that will be used to compute benchmark at
        redemption spread in case the bond is an australian FRN. Optional. No override
        is applied by default and price is computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("efpBenchmarkPrice")

    @efp_benchmark_price.setter
    def efp_benchmark_price(self, value):
        self._set_parameter("efpBenchmarkPrice", value)

    @property
    def efp_benchmark_ric(self):
        """
        RIC of EFP benchmark to override and that will be used as pricing analysis input
        to compute the bond price in case the bond is an australian FRN. Ric can be
        only "YTTc1" or "YTCc1". Optional. Default value is "YTTc1".
        :return: str
        """
        return self._get_parameter("efpBenchmarkRic")

    @efp_benchmark_ric.setter
    def efp_benchmark_ric(self, value):
        self._set_parameter("efpBenchmarkRic", value)

    @property
    def efp_benchmark_yield_percent(self):
        """
        Yield of EFP benchmark to override and that will be used to compute benchmark at
        redemption spread in case the bond is an australian FRN. Optional. No override
        is applied by default and yield is computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("efpBenchmarkYieldPercent")

    @efp_benchmark_yield_percent.setter
    def efp_benchmark_yield_percent(self, value):
        self._set_parameter("efpBenchmarkYieldPercent", value)

    @property
    def efp_spread_bp(self):
        """
        Spread of EFP benchmark to override and that will be used as pricing analysis
        input to compute the bond price in case the bond is an australian FRN. Optional.
        No override is applied by default. Note that only one pricing analysis input
        should be defined.
        :return: float
        """
        return self._get_parameter("efpSpreadBp")

    @efp_spread_bp.setter
    def efp_spread_bp(self, value):
        self._set_parameter("efpSpreadBp", value)

    @property
    def flat_credit_spread_bp(self):
        """
        Flat credit spread applied during pricing in basis points. Applicable when
        SpreadType = FlatSpread. Applicable for Convertible Bonds.
        :return: float
        """
        return self._get_parameter("flatCreditSpreadBp")

    @flat_credit_spread_bp.setter
    def flat_credit_spread_bp(self, value):
        self._set_parameter("flatCreditSpreadBp", value)

    @property
    def flat_credit_spread_tenor(self):
        """
        Flat credit spread tenor on credit curve used during pricing to source credit
        spread value. Applicable for Convertible Bonds.
        :return: str
        """
        return self._get_parameter("flatCreditSpreadTenor")

    @flat_credit_spread_tenor.setter
    def flat_credit_spread_tenor(self, value):
        self._set_parameter("flatCreditSpreadTenor", value)

    @property
    def fx_stock_correlation(self):
        """
        Correlation rate between underlying stock price and FX rate. Applicable for
        cross-currency Convertible Bonds.
        :return: float
        """
        return self._get_parameter("fxStockCorrelation")

    @fx_stock_correlation.setter
    def fx_stock_correlation(self, value):
        self._set_parameter("fxStockCorrelation", value)

    @property
    def fx_volatility_percent(self):
        """
        FX volatility rate percent. Applicable for cross-currency Convertible Bonds.
        :return: float
        """
        return self._get_parameter("fxVolatilityPercent")

    @fx_volatility_percent.setter
    def fx_volatility_percent(self, value):
        self._set_parameter("fxVolatilityPercent", value)

    @property
    def fx_volatility_tenor(self):
        """
        Tenor on FX volatility to source FX volatility Rate Percent. Applicable for
        cross-currency Convertible Bonds.
        :return: str
        """
        return self._get_parameter("fxVolatilityTenor")

    @fx_volatility_tenor.setter
    def fx_volatility_tenor(self, value):
        self._set_parameter("fxVolatilityTenor", value)

    @property
    def gov_country_benchmark_curve_price(self):
        """
        Price of government country benchmark to override and that will be used to
        compute user defined spread. Optional. No override is applied by default and
        price is computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("govCountryBenchmarkCurvePrice")

    @gov_country_benchmark_curve_price.setter
    def gov_country_benchmark_curve_price(self, value):
        self._set_parameter("govCountryBenchmarkCurvePrice", value)

    @property
    def gov_country_benchmark_curve_yield_percent(self):
        """
        Yield of government country benchmark to override and that will be used to
        compute government country spread. Optional. No override is applied by default
        and yield is computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("govCountryBenchmarkCurveYieldPercent")

    @gov_country_benchmark_curve_yield_percent.setter
    def gov_country_benchmark_curve_yield_percent(self, value):
        self._set_parameter("govCountryBenchmarkCurveYieldPercent", value)

    @property
    def gov_country_spread_bp(self):
        """
        Spread of government country benchmark to override and that will be used as
        pricing analysis input to compute the bond price. Optional. No override is
        applied by default. Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("govCountrySpreadBp")

    @gov_country_spread_bp.setter
    def gov_country_spread_bp(self, value):
        self._set_parameter("govCountrySpreadBp", value)

    @property
    def government_benchmark_curve_price(self):
        """
        price of government benchmark to override and that will be used to compute user
        defined spread. Optional. No override is applied by default and price is
        computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("governmentBenchmarkCurvePrice")

    @government_benchmark_curve_price.setter
    def government_benchmark_curve_price(self, value):
        self._set_parameter("governmentBenchmarkCurvePrice", value)

    @property
    def government_benchmark_curve_yield_percent(self):
        """
        Yield of government benchmark to override and that will be used to compute
        government spread. Optional. No override is applied by default and yield is
        computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("governmentBenchmarkCurveYieldPercent")

    @government_benchmark_curve_yield_percent.setter
    def government_benchmark_curve_yield_percent(self, value):
        self._set_parameter("governmentBenchmarkCurveYieldPercent", value)

    @property
    def government_spread_bp(self):
        """
        Spread of government benchmark to override and that will be used as pricing
        analysis input to compute the bond price. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("governmentSpreadBp")

    @government_spread_bp.setter
    def government_spread_bp(self, value):
        self._set_parameter("governmentSpreadBp", value)

    @property
    def issuer_benchmark_curve_yield_percent(self):
        """
        Yield of issuer benchmark to override and that will be used to compute issuer
        spread. Optional. No override is applied by default and yield is computed or
        retrieved from market data.
        :return: float
        """
        return self._get_parameter("issuerBenchmarkCurveYieldPercent")

    @issuer_benchmark_curve_yield_percent.setter
    def issuer_benchmark_curve_yield_percent(self, value):
        self._set_parameter("issuerBenchmarkCurveYieldPercent", value)

    @property
    def issuer_spread_bp(self):
        """
        Spread of issuer benchmark to override and that will be used as pricing analysis
        input to compute the bond price. This spread is computed is for coprorate bonds.
        Optional. No override is applied by default. Note that only one pricing anlysis
        input should be defined.
        :return: float
        """
        return self._get_parameter("issuerSpreadBp")

    @issuer_spread_bp.setter
    def issuer_spread_bp(self, value):
        self._set_parameter("issuerSpreadBp", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing. Optional. By default, the market_data_date
        date is the valuation_date or Today
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def market_value_in_deal_ccy(self):
        """
        Market value in deal currency. This field can be used to compute notionalAmount
        to apply to get this market value. Optional. By default the value is computed
        from notional amount. NotionalAmount field, market_value_in_deal_ccy field and
        market_value_in_report_ccy field cannot be set at defined at the same time.
        :return: float
        """
        return self._get_parameter("marketValueInDealCcy")

    @market_value_in_deal_ccy.setter
    def market_value_in_deal_ccy(self, value):
        self._set_parameter("marketValueInDealCcy", value)

    @property
    def market_value_in_report_ccy(self):
        """
        Market value in report currency. This field can be used to compute
        notionalAmount to apply to get this market value. Optional. By default the value
        is computed from notional amount. NotionalAmount field, market_value_in_deal_ccy
        field and market_value_in_report_ccy field cannot be set at defined at the same
        time.
        :return: float
        """
        return self._get_parameter("marketValueInReportCcy")

    @market_value_in_report_ccy.setter
    def market_value_in_report_ccy(self, value):
        self._set_parameter("marketValueInReportCcy", value)

    @property
    def net_price(self):
        """
        Net price to override and that will be used as pricing analysis input. Optional.
        No override is applied by default. Note that only one pricing anlysis input
        should be defined.
        :return: float
        """
        return self._get_parameter("netPrice")

    @net_price.setter
    def net_price(self, value):
        self._set_parameter("netPrice", value)

    @property
    def neutral_yield_percent(self):
        """
        Neutral Yield (expressed in percent) to override and that will be used as
        pricing analysis input. This is available only for floating rate notes.
        Optional. No override is applied by default. Note that only one pricing analysis
        input should be defined.
        :return: float
        """
        return self._get_parameter("neutralYieldPercent")

    @neutral_yield_percent.setter
    def neutral_yield_percent(self, value):
        self._set_parameter("neutralYieldPercent", value)

    @property
    def ois_zc_benchmark_curve_yield_percent(self):
        """
        Yield of OIS benchmark to override and that will be used to compute OIS spread.
        Optional. No override is applied by default and yield is computed or retrieved
        from market data.
        :return: float
        """
        return self._get_parameter("oisZcBenchmarkCurveYieldPercent")

    @ois_zc_benchmark_curve_yield_percent.setter
    def ois_zc_benchmark_curve_yield_percent(self, value):
        self._set_parameter("oisZcBenchmarkCurveYieldPercent", value)

    @property
    def ois_zc_spread_bp(self):
        """
        Yield of OIS benchmark to override and that will be used as pricing analysis
        input to compute the bond price. Optional. No override is applied by default.
        Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("oisZcSpreadBp")

    @ois_zc_spread_bp.setter
    def ois_zc_spread_bp(self, value):
        self._set_parameter("oisZcSpreadBp", value)

    @property
    def option_adjusted_spread_bp(self):
        """
        Option Adjusted Spread to override and that will be used as pricing analysis
        input to compute the bond price. Optional. No override is applied by default.
        Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("optionAdjustedSpreadBp")

    @option_adjusted_spread_bp.setter
    def option_adjusted_spread_bp(self, value):
        self._set_parameter("optionAdjustedSpreadBp", value)

    @property
    def price(self):
        """
        Price to override and that will be used as pricing analysis input. This price
        can be the clean price or dirty price depending on price type defined in bond
        structure. The currency of the price is the cash flow currency (that can be
        different to deal currency especially if "ComputeCashFlowWithReportCcy" flag has
        been set to true). Optional. No override is applied by default. Note that only
        one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("price")

    @price.setter
    def price(self, value):
        self._set_parameter("price", value)

    @property
    def quoted_price(self):
        """
        Quoted price to override and that will be used as pricing analysis input. Note
        that a quoted price can be a price, a yield, a discount margin, a spread,...
        depending on quotation type. The currency of the quoted price in case the bonnd
        is price-quoted or cash-quoted is the deal currency (that can be different to
        cash flow currency especially if "ComputeCashFlowWithReportCcy" flag has been
        set to true). Optional. No override is applied by default. Note that only one
        pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("quotedPrice")

    @quoted_price.setter
    def quoted_price(self, value):
        self._set_parameter("quotedPrice", value)

    @property
    def rating_benchmark_curve_yield_percent(self):
        """
        Yield of rating benchmark to override and that will be used to compute rating
        spread. Optional. No override is applied by default and yield is computed or
        retrieved from market data.
        :return: float
        """
        return self._get_parameter("ratingBenchmarkCurveYieldPercent")

    @rating_benchmark_curve_yield_percent.setter
    def rating_benchmark_curve_yield_percent(self, value):
        self._set_parameter("ratingBenchmarkCurveYieldPercent", value)

    @property
    def rating_spread_bp(self):
        """
        Spread of rating benchmark to override and that will be used as pricing analysis
        input to compute the bond price. Optional. No override is applied by default.
        Note that only one pricing anlysis input should be defined.
        :return: float
        """
        return self._get_parameter("ratingSpreadBp")

    @rating_spread_bp.setter
    def rating_spread_bp(self, value):
        self._set_parameter("ratingSpreadBp", value)

    @property
    def redemption_date(self):
        """
        Redemption date that defines the end date for yield and price computation. Used
        only if redemption date type is set to "RedemptionAtCustomDate"
        :return: str
        """
        return self._get_parameter("redemptionDate")

    @redemption_date.setter
    def redemption_date(self, value):
        self._set_parameter("redemptionDate", value)

    @property
    def sector_rating_benchmark_curve_yield_percent(self):
        """
        Yield of sector rating benchmark to override and that will be used to compute
        sector rating spread. Optional. No override is applied by default and yield is
        computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("sectorRatingBenchmarkCurveYieldPercent")

    @sector_rating_benchmark_curve_yield_percent.setter
    def sector_rating_benchmark_curve_yield_percent(self, value):
        self._set_parameter("sectorRatingBenchmarkCurveYieldPercent", value)

    @property
    def sector_rating_spread_bp(self):
        """
        Spread of sector rating benchmark to override and that will be used as pricing
        analysis input to compute the bond price. Optional. No override is applied by
        default. Note that only one pricing anlysis input should be defined.
        :return: float
        """
        return self._get_parameter("sectorRatingSpreadBp")

    @sector_rating_spread_bp.setter
    def sector_rating_spread_bp(self, value):
        self._set_parameter("sectorRatingSpreadBp", value)

    @property
    def settlement_convention(self):
        """
        Settlement convention for the bond. By default the rule is that valuation_date =
        trade_date + settlement_convention. Optional. By default use the settlement
        tenor defined in the bond structure. Only two parameters among
        "settlement_convention", "market_data_date" and "valuation_date" can be
        overriden at the same time.
        :return: str
        """
        return self._get_parameter("settlementConvention")

    @settlement_convention.setter
    def settlement_convention(self, value):
        self._set_parameter("settlementConvention", value)

    @property
    def simple_margin_bp(self):
        """
        Simple Margin basis points  to override and that will be used as pricing
        analysis input. Available only for Floating Rate Notes. Optional.No override is
        applied by default. Note that only one pricing anlysis input should be defined.
        :return: float
        """
        return self._get_parameter("simpleMarginBp")

    @simple_margin_bp.setter
    def simple_margin_bp(self, value):
        self._set_parameter("simpleMarginBp", value)

    @property
    def stock_borrow_rate_percent(self):
        """
        Underlying stock borrow rate percent. Applicable for Convertible Bonds.
        :return: float
        """
        return self._get_parameter("stockBorrowRatePercent")

    @stock_borrow_rate_percent.setter
    def stock_borrow_rate_percent(self, value):
        self._set_parameter("stockBorrowRatePercent", value)

    @property
    def stock_flat_volatility_percent(self):
        """
        Underlying stock volatility percent used for convertible pricing. Applicable
        when volatility_type = Flat Applicable for Convertible Bonds.
        :return: float
        """
        return self._get_parameter("stockFlatVolatilityPercent")

    @stock_flat_volatility_percent.setter
    def stock_flat_volatility_percent(self, value):
        self._set_parameter("stockFlatVolatilityPercent", value)

    @property
    def stock_flat_volatility_tenor(self):
        """
        Underlying Stock volatility tenor used during pricing to source volatility
        percent value. Applicable when volatility_type = Flat Applicable for Convertible
        Bonds.
        :return: str
        """
        return self._get_parameter("stockFlatVolatilityTenor")

    @stock_flat_volatility_tenor.setter
    def stock_flat_volatility_tenor(self, value):
        self._set_parameter("stockFlatVolatilityTenor", value)

    @property
    def stock_price_on_default(self):
        """
        Assumed stock price agreed in event of default. Applicable for Convertible
        Bonds.
        :return: float
        """
        return self._get_parameter("stockPriceOnDefault")

    @stock_price_on_default.setter
    def stock_price_on_default(self, value):
        self._set_parameter("stockPriceOnDefault", value)

    @property
    def strip_yield_percent(self):
        """
        Strip Yield (expressed in percent) to override and that will be used as pricing
        analysis input. Optional. No override is applied by default. Note that only one
        pricing anlysis input should be defined.
        :return: float
        """
        return self._get_parameter("stripYieldPercent")

    @strip_yield_percent.setter
    def strip_yield_percent(self, value):
        self._set_parameter("stripYieldPercent", value)

    @property
    def swap_benchmark_curve_yield_percent(self):
        """
        Yield of swap benchmark to override and that will be used to compute swap
        spread. Optional. No override is applied by default and yield is computed or
        retrieved from market data.
        :return: float
        """
        return self._get_parameter("swapBenchmarkCurveYieldPercent")

    @swap_benchmark_curve_yield_percent.setter
    def swap_benchmark_curve_yield_percent(self, value):
        self._set_parameter("swapBenchmarkCurveYieldPercent", value)

    @property
    def swap_spread_bp(self):
        """
        Spread of swap benchmark to override and that will be used as pricing analysis
        input to compute the bond price. Optional. No override is applied by default.
        Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("swapSpreadBp")

    @swap_spread_bp.setter
    def swap_spread_bp(self, value):
        self._set_parameter("swapSpreadBp", value)

    @property
    def tax_on_capital_gain_percent(self):
        """
        Tax Rate on capital gain expressed in percent. Optional. By default no tax is
        applied that means value is equal to 0.
        :return: float
        """
        return self._get_parameter("taxOnCapitalGainPercent")

    @tax_on_capital_gain_percent.setter
    def tax_on_capital_gain_percent(self, value):
        self._set_parameter("taxOnCapitalGainPercent", value)

    @property
    def tax_on_coupon_percent(self):
        """
        Tax Rate on Coupon  expressed in percent. Optional. By default no tax is applied
        that means value is equal to 0.
        :return: float
        """
        return self._get_parameter("taxOnCouponPercent")

    @tax_on_coupon_percent.setter
    def tax_on_coupon_percent(self, value):
        self._set_parameter("taxOnCouponPercent", value)

    @property
    def tax_on_price_percent(self):
        """
        Tax Rate on price  expressed in percent. Optional. By default no tax is applied
        that means value is equal to 0.
        :return: float
        """
        return self._get_parameter("taxOnPricePercent")

    @tax_on_price_percent.setter
    def tax_on_price_percent(self, value):
        self._set_parameter("taxOnPricePercent", value)

    @property
    def tax_on_yield_percent(self):
        """
        Tax Rate on Yield expressed in percent. Also named Tax on Yield Optional. By
        default no tax is applied that means value is equal to 0.
        :return: float
        """
        return self._get_parameter("taxOnYieldPercent")

    @tax_on_yield_percent.setter
    def tax_on_yield_percent(self, value):
        self._set_parameter("taxOnYieldPercent", value)

    @property
    def trade_date(self):
        """
        Trade date of the bond. The trade_date is used to compute the default
        valuation_date : By default the rule is that valuation_date = trade_date +
        settlement_convention. Optional. By default, it is equal to market_data_date.
        :return: str
        """
        return self._get_parameter("tradeDate")

    @trade_date.setter
    def trade_date(self, value):
        self._set_parameter("tradeDate", value)

    @property
    def use_settlement_date_from_quote(self):
        """
        Specify whether to use the settlment date of the quote or the one computed from
        the MarketData Date
        :return: bool
        """
        return self._get_parameter("useSettlementDateFromQuote")

    @use_settlement_date_from_quote.setter
    def use_settlement_date_from_quote(self, value):
        self._set_parameter("useSettlementDateFromQuote", value)

    @property
    def user_defined_benchmark_price(self):
        """
        Price of user defined instrument to override and that will be used to compute
        user defined spread. Optional. No override is applied by default and price is
        computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("userDefinedBenchmarkPrice")

    @user_defined_benchmark_price.setter
    def user_defined_benchmark_price(self, value):
        self._set_parameter("userDefinedBenchmarkPrice", value)

    @property
    def user_defined_benchmark_yield_percent(self):
        """
        Yield of user defined instrument to override and that will be used to compute
        user defined spread. Optional. No override is applied by default and yield is
        computed or retrieved from market data.
        :return: float
        """
        return self._get_parameter("userDefinedBenchmarkYieldPercent")

    @user_defined_benchmark_yield_percent.setter
    def user_defined_benchmark_yield_percent(self, value):
        self._set_parameter("userDefinedBenchmarkYieldPercent", value)

    @property
    def user_defined_spread_bp(self):
        """
        Spread of user defined instrument to override and that will be used as pricing
        analysis input to compute the bond price. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("userDefinedSpreadBp")

    @user_defined_spread_bp.setter
    def user_defined_spread_bp(self, value):
        self._set_parameter("userDefinedSpreadBp", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.  Optional. If not set the valuation date is
        equal to market_data_date or Today. For assets that contains a
        settlement_convention, the default valuation date  is equal to the
        settlementdate of the Asset that is usually the
        trade_date+settlement_convention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)

    @property
    def yield_percent(self):
        """
        Yield (expressed in percent) to override and that will be used as pricing
        analysis input. Optional. No override is applied by default. Note that only one
        pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("yieldPercent")

    @yield_percent.setter
    def yield_percent(self, value):
        self._set_parameter("yieldPercent", value)

    @property
    def z_spread_bp(self):
        """
        ZSpread to override and that will be used as pricing analysis input to compute
        the bond price. Optional. No override is applied by default. Note that only one
        pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("zSpreadBp")

    @z_spread_bp.setter
    def z_spread_bp(self, value):
        self._set_parameter("zSpreadBp", value)



# coding: utf8

__all__ = ["BarrierMode"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class BarrierMode(Enum):
    AMERICAN = "American"
    EARLY_END_WINDOW = "EarlyEndWindow"
    EUROPEAN = "European"
    FORWARD_START_WINDOW = "ForwardStartWindow"
    UNDEFINED = "Undefined"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(BarrierMode, _BARRIER_MODE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_BARRIER_MODE_VALUES_IN_LOWER_BY_BARRIER_MODE, some)


_BARRIER_MODE_VALUES = tuple(t.value for t in BarrierMode)
_BARRIER_MODE_VALUES_IN_LOWER_BY_BARRIER_MODE = {
    name.lower(): item for name, item in BarrierMode.__members__.items()
}



# coding: utf8

from typing import Optional

from .._base import BarrierDefinition
from .._enums import (
    BarrierStyle,
    InOrOut,
    UpOrDown,
)


class EtiBarrierDefinition(BarrierDefinition):
    """
    Parameters
    ----------
    barrier_style : BarrierStyle, optional

    in_or_out : InOrOut, optional

    up_or_down : UpOrDown, optional

    level : float, optional

    """

    def __init__(
        self,
        barrier_style: Optional[BarrierStyle] = None,
        in_or_out: Optional[InOrOut] = None,
        up_or_down: Optional[UpOrDown] = None,
        level: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.barrier_style = barrier_style
        self.in_or_out = in_or_out
        self.up_or_down = up_or_down
        self.level = level

    @property
    def barrier_style(self):
        """
        :return: enum BarrierStyle
        """
        return self._get_enum_parameter(BarrierStyle, "barrierStyle")

    @barrier_style.setter
    def barrier_style(self, value):
        self._set_enum_parameter(BarrierStyle, "barrierStyle", value)

    @property
    def in_or_out(self):
        """
        :return: enum InOrOut
        """
        return self._get_enum_parameter(InOrOut, "inOrOut")

    @in_or_out.setter
    def in_or_out(self, value):
        self._set_enum_parameter(InOrOut, "inOrOut", value)

    @property
    def up_or_down(self):
        """
        :return: enum UpOrDown
        """
        return self._get_enum_parameter(UpOrDown, "upOrDown")

    @up_or_down.setter
    def up_or_down(self, value):
        self._set_enum_parameter(UpOrDown, "upOrDown", value)

    @property
    def level(self):
        """
        :return: float
        """
        return self._get_parameter("level")

    @level.setter
    def level(self, value):
        self._set_parameter("level", value)



import abc
from itertools import product
from typing import List, Optional, Any, Set

import numpy as np
import pandas as pd
from pandas import DataFrame

from refinitiv.data._tools._dataframe import (
    convert_str_to_datetime,
    convert_items_df_to_datetime,
)


class DFBuilder(abc.ABC):
    auto_headers: Optional[List] = None

    @abc.abstractmethod
    def get_headers(self, content_data: dict) -> List[dict]:
        pass

    def get_date_idxs(self, columns: List[str]) -> Set[int]:
        return {idx for idx, col in enumerate(columns) if "date" in col.lower()}

    def build_index(
        self, content_data: dict, use_field_names_in_headers: bool = False, **kwargs
    ) -> pd.DataFrame:

        key = "name" if use_field_names_in_headers else "title"

        headers = self.get_headers(content_data)
        columns = [header[key] for header in headers]
        date_idxs = self.get_date_idxs(columns)

        data = []
        fields_list = content_data.get("data", [])
        for fields in fields_list:
            fields = list(fields)

            for idx, item in enumerate(fields):
                if item is None:
                    fields[idx] = pd.NA

            for idx in date_idxs:
                fields[idx] = convert_str_to_datetime(fields[idx])

            data.append(fields)

        df = pd.DataFrame(data=data, columns=columns)
        return df

    def build_date_as_index(
        self, content_data: dict, use_field_names_in_headers: bool = False, **kwargs
    ) -> pd.DataFrame:
        key = "name" if use_field_names_in_headers else "title"

        headers = self.get_headers(content_data)
        columns = [header[key] for header in headers]

        inst_header, date_header = self.auto_headers
        inst_header = inst_header[key]
        date_header = date_header[key]

        data = []
        index = []
        inst_idx = columns.index(inst_header)
        date_idx = columns.index(date_header)
        columns.pop(date_idx)
        columns.pop(inst_idx)
        num_columns = len(columns)

        date_idxs = self.get_date_idxs(columns)
        fields_by_inst_by_date = {}
        fields_list = content_data.get("data", [])
        unique_insts = list(dict.fromkeys(fields[inst_idx] for fields in fields_list))
        num_unique_insts = len(unique_insts)
        for fields in fields_list:
            fields = list(fields)
            date_str = fields[date_idx]

            if not date_str:
                continue

            inst = fields[inst_idx]
            fields.pop(date_idx)
            fields.pop(inst_idx)
            is_add = True

            for idx, item in enumerate(fields):
                if item is None:
                    fields[idx] = pd.NA

            for idx in date_idxs:
                fields[idx] = convert_str_to_datetime(fields[idx])

            if num_unique_insts > 1:
                total = num_unique_insts * num_columns
                template = [pd.NA for _ in range(total)]

                idx = unique_insts.index(inst)
                right_idx = idx * num_columns + num_columns
                left_idx = idx * num_columns
                for item, idx in zip(fields, range(left_idx, right_idx)):
                    template[idx] = item
                fields = template

                fields_by_inst = fields_by_inst_by_date.get(date_str)
                if fields_by_inst and inst not in fields_by_inst.keys():
                    is_add = False
                    idx = max(
                        unique_insts.index(inst) for inst in fields_by_inst.keys()
                    )
                    cache_inst = unique_insts[idx]
                    cache_fields = fields_by_inst[cache_inst]
                    idx = unique_insts.index(inst)
                    cache_idx = unique_insts.index(cache_inst)

                    left_idx = cache_idx * num_columns + num_columns
                    right_idx = idx * num_columns + num_columns

                    for idx in range(left_idx, right_idx):
                        cache_fields[idx] = fields[idx]

                    fields_by_inst[inst] = cache_fields
                    fields = cache_fields

                else:
                    not fields_by_inst and fields_by_inst_by_date.setdefault(
                        date_str, {inst: fields}
                    )
                    date = convert_str_to_datetime(date_str)
                    index.append(date)

            else:
                date = convert_str_to_datetime(date_str)
                index.append(date)

            is_add and data.append(fields)

        if num_unique_insts > 1:
            columns = pd.MultiIndex.from_tuples(product(unique_insts, columns))

        elif num_unique_insts == 1:
            columns = pd.Index(data=columns, name=unique_insts.pop())

        index = pd.Index(data=index, name=date_header)
        df = pd.DataFrame(data=data, columns=columns, index=index)
        df.sort_index(ascending=False, inplace=True)
        return df


class DFBuilderRDP(DFBuilder):
    """
    {
        "links": {"count": 2},
        "variability": "",
        "universe": [
            {
                "Instrument": "GOOG.O",
                "Company Common Name": "Alphabet Inc",
                "Organization PermID": "5030853586",
                "Reporting Currency": "USD",
            }
        ],
        "data": [
            ["GOOG.O", "2022-01-26T00:00:00", "USD", None],
            ["GOOG.O", "2020-12-31T00:00:00", None, "2020-12-31T00:00:00"],
        ],
        "messages": {
            "codes": [[-1, -1, -1, -2], [-1, -1, -2, -1]],
            "descriptions": [
                {"code": -2, "description": "empty"},
                {"code": -1, "description": "ok"},
            ],
        },
        "headers": [
            {
                "name": "instrument",
                "title": "Instrument",
                "type": "string",
                "description": "The requested Instrument as defined by the user.",
            },
            {
                "name": "date",
                "title": "Date",
                "type": "datetime",
                "description": "Date associated with the returned data.",
            },
            {
                "name": "TR.RevenueMean",
                "title": "Currency",
                "type": "string",
                "description": "The statistical average of all broker ...",
            },
            {
                "name": "TR.Revenue",
                "title": "Date",
                "type": "datetime",
                "description": "Is used for industrial and utility companies. ...",
            },
        ],
    }
    """

    auto_headers = [
        {"name": "instrument", "title": "Instrument"},
        {"name": "date", "title": "Date"},
    ]

    def get_headers(self, content_data) -> List[dict]:
        return content_data.get("headers", [])


class DFBuilderUDF(DFBuilder):
    """
    {
        "columnHeadersCount": 1,
        "data": [
            ["GOOG.O", "2022-01-26T00:00:00Z", "USD", ""],
            ["GOOG.O", "2020-12-31T00:00:00Z", "", "2020-12-31T00:00:00Z"],
        ],
        "headerOrientation": "horizontal",
        "headers": [
            [
                {"displayName": "Instrument"},
                {"displayName": "Date"},
                {"displayName": "Currency", "field": "TR.REVENUEMEAN.currency"},
                {"displayName": "Date", "field": "TR.REVENUE.DATE"},
            ]
        ],
        "rowHeadersCount": 2,
        "totalColumnsCount": 4,
        "totalRowsCount": 3,
    }
    """

    auto_headers = [
        {"name": "Instrument", "title": "Instrument"},
        {"name": "Date", "title": "Date"},
    ]

    def get_headers(self, content_data) -> List[dict]:
        headers = content_data["headers"]
        headers = headers[0]
        return [
            {
                "name": header.get("field") or header.get("displayName"),
                "title": header.get("displayName"),
            }
            for header in headers
        ]


def default_build_df(raw: Any, **kwargs) -> pd.DataFrame:
    df = pd.DataFrame(raw)
    return df


def build_empty_df(*args, **kwargs) -> pd.DataFrame:
    return pd.DataFrame()


dfbuilder_rdp = DFBuilderRDP()
dfbuilder_udf = DFBuilderUDF()


def historical_build_df(raw, **kwargs):
    data = raw.get("data")

    if not data:
        return DataFrame()

    headers = raw.get("headers", [])
    fields = [field["name"] for field in headers]

    if not fields:
        return DataFrame()

    field_timestamp = None
    if "DATE_TIME" in fields:
        field_timestamp = "DATE_TIME"
    elif "DATE" in fields:
        field_timestamp = "DATE"

    if not field_timestamp:
        return DataFrame()

    timestamp_index = fields.index(field_timestamp)
    # remove timestamp from fields (timestamp is used as index for dataframe)
    fields.pop(timestamp_index)

    # remove timestamp from array
    timestamps = [l.pop(timestamp_index) for l in data]

    df = DataFrame(
        np.array(data),
        columns=fields,
        index=convert_items_df_to_datetime(timestamps),
    )
    if not df.empty:
        df.fillna(pd.NA, inplace=True)
        df = df.convert_dtypes()
        df.sort_index(inplace=True)

    return df



import config as ext_config_mod

from typing import Union

from ._desktop_session import DesktopSession
from ._session_type import SessionType
from ._platform_session import PlatformSession
from ... import _configure as configure
from .grant_password import GrantPassword

CANNOT_FIND_APP_KEY = "Can't find 'app-key' in config object."

DEPLOYED_PLATFORM_URL_CONFIG_KEY = "realtime-distribution-system.url"
DEPLOYED_PLATFORM_USER_CONFIG_KEY = "realtime-distribution-system.dacs.username"
DEPLOYED_PLATFORM_DACS_POS_CONFIG_KEY = "realtime-distribution-system.dacs.position"
DEPLOYED_PLATFORM_DACS_ID_CONFIG_KEY = (
    "realtime-distribution-system.dacs.application-id"
)

PLATFORM_ARGUMENT_KEY_TO_CONFIG_KEY = {
    "app_key": "app-key",
    "signon_control": "signon_control",
    "deployed_platform_host": DEPLOYED_PLATFORM_URL_CONFIG_KEY,
    "deployed_platform_username": DEPLOYED_PLATFORM_USER_CONFIG_KEY,
    "dacs_position": DEPLOYED_PLATFORM_DACS_POS_CONFIG_KEY,
    "dacs_application_id": DEPLOYED_PLATFORM_DACS_ID_CONFIG_KEY,
    "realtime_distribution_system_url": DEPLOYED_PLATFORM_URL_CONFIG_KEY,
    "auto_reconnect": "auto-reconnect",
    "server_mode": "server-mode",
    "base_url": "base-url",
    "auth_url": "auth.url",
    "auth_authorize": "auth.authorize",
    "auth_token": "auth.token",
}

DESKTOP_ARGUMENT_KEY_TO_CONFIG_KEY = {
    "app_key": "app-key",
    "token": "token",
    "dacs_position": DEPLOYED_PLATFORM_DACS_POS_CONFIG_KEY,
    "dacs_application_id": DEPLOYED_PLATFORM_DACS_ID_CONFIG_KEY,
    "base_url": "base-url",
    "platform_path_rdp": "platform-paths.rdp",
    "platform_path_udf": "platform-paths.udf",
    "handshake_url": "handshake-url",
}

session_class_by_session_type = {
    SessionType.DESKTOP: DesktopSession,
    SessionType.PLATFORM: PlatformSession,
}

session_alias_by_session_type = {
    "platform": SessionType.PLATFORM,
    "desktop": SessionType.DESKTOP,
}


def get_session_type(value: Union[SessionType, str]):
    session_type = None

    if isinstance(value, SessionType):
        session_type = value

    elif isinstance(value, str):
        session_type = session_alias_by_session_type.get(value)

    if not session_type:
        raise ValueError(f"Cannot get session type by value:{value}")

    return session_type


def _retrieve_values_from_config(config, session_type, grant):
    arguments = {}
    if session_type == SessionType.PLATFORM:
        username = config.get("username")
        password = config.get("password")
        arguments["grant"] = grant or GrantPassword(username, password)

        for argument, value in PLATFORM_ARGUMENT_KEY_TO_CONFIG_KEY.items():
            arguments[argument] = config.get(value)
    else:

        for argument, value in DESKTOP_ARGUMENT_KEY_TO_CONFIG_KEY.items():
            arguments[argument] = config.get(value)

    return arguments


def _validate_platform_session_arguments(
    app_key=None,
    grant=None,
    deployed_platform_host=None,
    deployed_platform_username=None,
):
    if app_key == "":
        raise AttributeError(CANNOT_FIND_APP_KEY)

    if app_key is None:
        raise AttributeError(
            "Please, set app-key in [session.platform.default] section of "
            "the config file or provide 'app-key' attribute to the definition."
        )

    if not grant.is_valid() and (
        not deployed_platform_host or not deployed_platform_username
    ):
        raise AttributeError(
            "To create platform session, please provide 'grant' attribute to the "
            "definition or set 'username' and 'password' in the config file. "
            "To create deployed session, please provide 'deployed_platform_host' "
            "and 'deployed_platform_username' to the definition or the config file."
        )


def _make_platform_session_provider_by_arguments(
    session_name,
    app_key=None,
    signon_control=None,
    deployed_platform_host=None,
    deployed_platform_username=None,
    dacs_position=None,
    dacs_application_id=None,
    grant=None,
):
    session_config = configure.get(configure.keys.platform_session(session_name), {})
    default_session_config = configure.get(
        configure.keys.platform_session("default"), {}
    )

    if isinstance(session_config, dict) and len(session_config) == 0:
        raise ValueError(
            f"Session name: {session_name} is invalid or session_name object is empty"
        )

    arguments = {
        "app-key": app_key,
        "signon_control": signon_control,
        DEPLOYED_PLATFORM_URL_CONFIG_KEY: deployed_platform_host,
        DEPLOYED_PLATFORM_USER_CONFIG_KEY: deployed_platform_username,
        DEPLOYED_PLATFORM_DACS_POS_CONFIG_KEY: dacs_position,
        DEPLOYED_PLATFORM_DACS_ID_CONFIG_KEY: dacs_application_id,
    }
    filtered_arguments = {
        key: value for key, value in arguments.items() if value is not None
    }

    merged_config = ext_config_mod.ConfigurationSet(
        ext_config_mod.config_from_dict(filtered_arguments),
        session_config,
        default_session_config,
    )

    app_key = merged_config.get("app-key")
    username = merged_config.get("username")
    password = merged_config.get("password")
    deployed_platform_host = merged_config.get(DEPLOYED_PLATFORM_URL_CONFIG_KEY)
    deployed_platform_username = merged_config.get(DEPLOYED_PLATFORM_USER_CONFIG_KEY)
    grant = grant or GrantPassword(username, password)

    _validate_platform_session_arguments(
        app_key=app_key,
        grant=grant,
        deployed_platform_host=deployed_platform_host,
        deployed_platform_username=deployed_platform_username,
    )

    return make_session_provider(
        SessionType.PLATFORM, merged_config, grant, session_name
    )


def _validate_desktop_session_app_key(
    session_config=None, app_key=None, session_name=None
):
    if not app_key:
        raise AttributeError(CANNOT_FIND_APP_KEY)

    if not session_config and not app_key:
        raise ValueError(
            f"Can't get config by name: {session_name}. Please check config name or provide app_key"
        )


def _make_desktop_session_provider_by_arguments(session_name, app_key=None):
    session_config = configure.get(configure.keys.desktop_session(session_name), {})
    default_session_config = configure.get(
        configure.keys.desktop_session("workspace"), {}
    )

    if isinstance(session_config, dict) and len(session_config) == 0:
        raise ValueError(
            f"Session name: {session_name} is invalid or session_name object is empty"
        )

    arguments = {"app-key": app_key}
    filtered_arguments = {
        key: value for key, value in arguments.items() if value is not None
    }

    merged_config = ext_config_mod.ConfigurationSet(
        ext_config_mod.config_from_dict(filtered_arguments),
        session_config,
        default_session_config,
    )

    _validate_desktop_session_app_key(
        session_config=session_config,
        app_key=merged_config.get("app-key"),
        session_name=session_name,
    )

    return make_session_provider(
        SessionType.DESKTOP, merged_config, session_name=session_name
    )


def _validate_session_arguments(
    app_key, session_type, grant, deployed_platform_host, deployed_platform_username
):
    if app_key is None:
        raise AttributeError(CANNOT_FIND_APP_KEY)

    if (
        session_type == SessionType.PLATFORM
        and not grant.is_valid()
        and (not deployed_platform_host or not deployed_platform_username)
    ):
        raise AttributeError(
            "Please provide 'grant' attribute or set 'username' and 'password' in config. "
            "Or provide deployed parameters in config file to create deployed platform session"
        )


def _make_session_provider_by_arguments(session_name):
    from ._session_definition import (
        _get_session_type_and_name,
        _retrieve_config_and_set_type,
    )

    config_path = session_name  # can't rename argument, because public API, make it right at least in function body
    session_name, session_type = _get_session_type_and_name(config_path)
    session_config = _retrieve_config_and_set_type(session_name, session_type)

    if session_type == SessionType.DESKTOP:
        default_session_config = configure.get(
            configure.keys.desktop_session("workspace"), {}
        )
    else:
        default_session_config = configure.get(
            configure.keys.platform_session("default"), {}
        )

    merged_config = ext_config_mod.ConfigurationSet(
        session_config, default_session_config
    )
    app_key = merged_config.get("app-key")
    deployed_platform_host = merged_config.get(DEPLOYED_PLATFORM_URL_CONFIG_KEY)
    deployed_platform_username = merged_config.get(DEPLOYED_PLATFORM_USER_CONFIG_KEY)

    grant = None
    if session_type == SessionType.PLATFORM:
        grant = GrantPassword(
            merged_config.get("username"), merged_config.get("password")
        )

    _validate_session_arguments(
        app_key=app_key,
        session_type=session_type,
        grant=grant,
        deployed_platform_host=deployed_platform_host,
        deployed_platform_username=deployed_platform_username,
    )

    return make_session_provider(session_type, merged_config, grant, session_name)


def make_session_provider(
    session_type, config=None, grant=None, session_name="default"
):
    config = config or {}
    session_class = session_class_by_session_type.get(session_type)
    if not session_class:
        raise ValueError(f"Cannot find session class by session type {session_type}")
    session_class = get_session_class(session_type)

    def session_provider():
        params = _retrieve_values_from_config(config, session_type, grant)
        params["name"] = session_name
        sessions_inst = session_class(**params)
        sessions_inst.debug(f" + Session created {sessions_inst}")
        return sessions_inst

    return session_provider


def get_session_class(session_type):
    session_class = session_class_by_session_type.get(session_type, None)

    if session_class is None:
        raise ValueError(f"Cannot find session class by session type {session_type}")

    return session_class



__all__ = (
    "AverageType",
    "BarrierMode",
    "BarrierStyle",
    "BidAskMid",
    "BinaryType",
    "BuySell",
    "CallPut",
    "DayWeight",
    "Definition",
    "DoubleBinaryType",
    "EtiBarrierDefinition",
    "EtiBinaryDefinition",
    "EtiCbbcDefinition",
    "EtiDoubleBarriersDefinition",
    "EtiFixingInfo",
    "EtiUnderlyingDefinition",
    "ExerciseStyle",
    "FixingFrequency",
    "FxAverageInfo",
    "FxBarrierDefinition",
    "FxBinaryDefinition",
    "FxBinaryType",
    "FxDoubleBarrierDefinition",
    "FxDoubleBarrierInfo",
    "FxDoubleBinaryDefinition",
    "FxDualCurrencyDefinition",
    "FxForwardStart",
    "FxSwapCalculationMethod",
    "FxUnderlyingDefinition",
    "InOrOut",
    "InterpolationWeight",
    "OptionVolatilityType",
    "PayoutScaling",
    "PriceSide",
    "PricingModelType",
    "PricingParameters",
    "SettlementType",
    "Status",
    "TimeStamp",
    "UnderlyingType",
    "UpOrDown",
    "VolatilityModel",
    "VolatilityType",
)

from ..._models import DayWeight

from ._definition import Definition
from ._enums import (
    AverageType,
    BarrierMode,
    BarrierStyle,
    BinaryType,
    BuySell,
    CallPut,
    DoubleBinaryType,
    ExerciseStyle,
    FixingFrequency,
    FxBinaryType,
    FxSwapCalculationMethod,
    InOrOut,
    OptionVolatilityType,
    PriceSide,
    PricingModelType,
    SettlementType,
    Status,
    TimeStamp,
    UnderlyingType,
    UpOrDown,
    VolatilityModel,
    VolatilityType,
)
from ._eti import (
    EtiBarrierDefinition,
    EtiBinaryDefinition,
    EtiCbbcDefinition,
    EtiDoubleBarriersDefinition,
    EtiFixingInfo,
    EtiUnderlyingDefinition,
)
from ._fx import (
    FxAverageInfo,
    FxBarrierDefinition,
    FxBinaryDefinition,
    FxDoubleBarrierDefinition,
    FxDoubleBarrierInfo,
    FxDoubleBinaryDefinition,
    FxDualCurrencyDefinition,
    FxForwardStart,
    FxUnderlyingDefinition,
)
from ._models import (
    BidAskMid,
    InterpolationWeight,
    PayoutScaling,
)
from ._option_pricing_parameters import PricingParameters
from ..._models import DayWeight
